from whad.zigbee.stack.apl.zcl import ZCLClientCluster
from whad.zigbee.stack.apl.zcl.clusters import *
"""
This module defines classes and exceptions related to ZigBee node's Endpoints.
"""

class ClusterNotAvailable(Exception):
    """
    Exception raised when a cluster is not available.
    """
    pass

class Endpoint:
    def __init__(self, number, node):
        self.__number = number
        self.__node = node
        self.__descriptor = None

    @property
    def descriptor(self):
        """
        Property returning the descriptor associated to this endpoint.
        """

        if (
            self.__descriptor is None and
            self.__node.network.is_associated() and
            self.__node.network.is_authorized()
        ):
            self.__descriptor = self.stack.get_layer('apl').get_application_by_name("zdo").device_and_service_discovery.get_simple_descriptor(
                self.__node.address,
                self.__number
            )
        return self.__descriptor

    @property
    def stack(self):
        """
        Property returning the stack instance linked to this endpoint.
        """
        if self.__node is None:
            return None
        return self.__node.stack

    @property
    def node(self):
        """
        Property returning of this endpoint.
        """
        return self.__node


    @property
    def profile_id(self):
        """
        Property returning the profile ID of this endpoint.
        """
        return self.descriptor.profile_identifier

    @property
    def device_id(self):
        """
        Property indicating the device ID of this endpoint.
        """
        return self.descriptor.device_identifier

    @property
    def input_clusters(self):
        """
        Property returning the input clusters of this endpoint.
        """
        return self.descriptor.input_clusters

    @property
    def output_clusters(self):
        """
        Property returning the output clusters of this endpoint.
        """
        return self.descriptor.output_clusters

    def attach_to_input_cluster(self, cluster_id, application=None):
        """
        Attach the endpoint to an input cluster in the stack, according
        to the cluster id.
        """
        # Check if the cluster number is in the endpoint's input cluster list
        if cluster_id not in self.descriptor.input_clusters:
            raise ClusterNotAvailable

        # Select the first application matching the profile
        selected_application = None
        for application in self.stack.get_layer('apl').get_applications():
            if application.profile_id == self.profile_id:
                selected_application = application
                break

        # Check if a matching client cluster exists in our application
        selected_cluster = None
        for cluster in selected_application.output_clusters:
            if cluster.cluster_id == cluster_id:
                selected_cluster = cluster
                break

        if selected_cluster is None:
            # Only works with ZCL, we should adapt it for other clusters
            for candidate_cluster_class in ZCLClientCluster.child_clusters():
                if candidate_cluster_class().cluster_id == cluster_id:
                    selected_cluster = candidate_cluster_class()
                    break

        if selected_cluster is not None:
            selected_application.add_output_cluster(selected_cluster)
            selected_cluster.connect(self.node.address, self.number)

        return selected_cluster


    @property
    def number(self):
        return self.__number

    def __repr__(self):
        return "Endpoint(#%d, input_clusters=%s, output_clusters=%s)" % (self.number, str(self.input_clusters), str(self.output_clusters))



"""Bluetooth Low Energy mocks
"""
from .device import EmulatedDevice
from .scan import DeviceScan
from .central import CentralMock

__all__ = [
    "CentralMock",
    "DeviceScan",
    "EmulatedDevice",
]



class ZigbeeNetwork:
    def __init__(self, beacon):
        self.dot15d4_pan_network = beacon.pan_descriptor
        self.extended_pan_id = beacon.extended_pan_id
        self.channel = self.dot15d4_pan_network.channel
        self.stack_profile = beacon.stack_profile
        self.zigbee_version = beacon.nwkc_protocol_version
        self.beacon_order = self.dot15d4_pan_network.beacon_order
        self.superframe_order = self.dot15d4_pan_network.superframe_order
        self.joining_permit = self.dot15d4_pan_network.assoc_permit
        self.router_capacity = bool(beacon.router_capacity)
        self.end_device_capacity = bool(beacon.end_device_capacity)

    def __eq__(self, other):
        return self.extended_pan_id == other.extended_pan_id

    def __repr__(self):
        return ("ZigbeeNetwork("+
                "pan_id=" + hex(self.dot15d4_pan_network.coord_pan_id)+", "
                "extended_pan_id=" + hex(self.extended_pan_id)+", "
                "pan_coord=" + (hex(self.dot15d4_pan_network.coord_addr) if isinstance(self.dot15d4_pan_network.coord_addr,int) else self.dot15d4_pan_network.coord_addr) + ", "
                "channel=" + str(self.channel)+", "
                "joining=" + ("permitted" if self.joining_permit else "forbidden") +", " +
                "router_capacity=" + ("yes" if self.router_capacity else "no") +", " +
                "end_device_capacity=" + ("yes" if self.end_device_capacity else "no") +
            ")"
        )



from enum import IntEnum
"""
Constants implemented as Enum for Zigbee APS layer.
"""

class APSSourceAddressMode(IntEnum):
    """
    Enum representing the different APS source address modes supported by Zigbee.
    """
    SHORT_ADDRESS_SRC_ENDPOINT_PRESENT = 0x02
    EXTENDED_ADDRESS_SRC_ENDPOINT_PRESENT = 0x03
    EXTENDED_ADDRESS_SRC_ENDPOINT_NOT_PRESENT = 0x04


class APSDestinationAddressMode(IntEnum):
    """
    Enum representing the different APS destination address modes supported by Zigbee.
    """
    DST_ADDRESS_AND_DST_ENDPOINT_NOT_PRESENT = 0x00
    SHORT_GROUP_ADDRESS_DST_ENDPOINT_NOT_PRESENT = 0x01
    SHORT_ADDRESS_DST_ENDPOINT_PRESENT = 0x02
    EXTENDED_ADDRESS_DST_ENDPOINT_PRESENT = 0x03
    EXTENDED_ADDRESS_DST_ENDPOINT_NOT_PRESENT = 0x04


class APSKeyAttribute(IntEnum):
    """
    Enum storing the different possible values for the APS key attribute.
    """
    PROVISIONAL_KEY = 0
    UNVERIFIED_KEY = 1
    VERIFIED_KEY = 2

class APSKeyType(IntEnum):
    """
    Enum storing the different possible values for the APS key type attribute.
    """
    STANDARD_NETWORK_KEY = 1
    APPLICATION_LINK_KEY = 3
    TRUST_CENTER_LINK_KEY = 4

class APSSecurityStatus(IntEnum):
    """
    Enum storing the different possible values for the APS security status.
    """
    UNSECURED = 0
    SECURED_NWK_KEY = 1
    SECURED_LINK_KEY = 2



"""
BLE GATT Service Model
======================
"""
from abc import abstractmethod, ABC
import logging
from typing import Optional, Type, Iterator, Union
from struct import pack
from threading import Thread

from whad.ble.profile.attribute import (
    Attribute, UUID, get_uuid_alias, InvalidUUIDException
)
from whad.ble.profile.characteristic import Characteristic
from whad.ble.utils.clues import CluesDb

logger = logging.getLogger(__name__)

class Service(Attribute):
    """GATT service attribute
    """

    def __init__(self, type_uuid: UUID, uuid: UUID, handle: int = 0, end_handle: int = 0, **children):
        """Instantiate a new service attribute with the specified handle.

        :param type_uuid: Attribute type UUID
        :type  type_uuid: UUID
        :param uuid: Service UUID
        :type  uuid: UUID
        :param handle: Service handle
        :type  handle: int
        :param end_handle: Handle of last attribute in the service
        :type  end_handle: int
        :param children: List of children attributes to add into this service
        :type  children: dict, optional
        """
        # Create our base attribute with type UUID, handle and value.
        super().__init__(uuid=type_uuid,handle=handle, value=uuid.to_bytes())

        # Set service UUID
        self.__service_uuid = uuid

        # Characteristics contained in service.
        self.__characteristics = []

        # Secondary services included in this service.
        self.__included_services = []

        # If handle is not 0, then we are not in template mode and we set this service without considering any
        # template characteristics or included services.
        if handle > 0:
            # Set characteristic end handle
            self.handle = handle
            self.__end_handle = end_handle
        else:
            # Template mode, we enumerate the class attributes to find template characteristics.
            self.__end_handle = 0
            for name, obj in self.inspect(Characteristic, Service):
                if isinstance(obj, Characteristic):
                    # For each template characteristic, we build it and replace the template object
                    # with this new one (but the template object is still in class definition).
                    charac = obj.build()
                    charac.service = self
                    charac.alias = name
                    self.add_characteristic(charac)
                    setattr(self, name, charac)
                elif isinstance(obj, Service):
                    service = obj.build()
                    self.add_included_service(IncludeService(service.uuid))

            # Add defined children (characteristic & include service)
            for name, obj in children.items():
                if isinstance(obj, Characteristic):
                    charac = obj.build()
                    charac.service = self
                    charac.alias = name
                    self.add_characteristic(charac)
                    setattr(self, name, charac)
                elif isinstance(obj, Service):
                    service = obj.build()
                    self.add_included_service(IncludeService(service.uuid))

    @property
    def uuid(self) -> UUID:
        """Service UUID
        """
        return self.__service_uuid

    @Attribute.handle.setter
    def handle(self, new_handle: int):
        """Overwrite `Attribute` handle setter.
        """
        # Update service handle
        Attribute.handle.fset(self, new_handle)

        # Update the underlying characteristics
        char_handle = new_handle
        for characteristic in self.__characteristics:
            characteristic.handle = char_handle + 1
            char_handle = characteristic.end_handle

        # Update service end_handle value
        self.__end_handle = char_handle

    @property
    def end_handle(self) -> int:
        """End handle
        """
        return self.__end_handle

    @end_handle.setter
    def end_handle(self, value: int):
        """Set end handle
        """
        self.__end_handle = value

    @property
    def name(self) -> str:
        """Readable service name

        :return: Service name as readable text, if either defined in the Bluetooth
                 specification or known from DarkMentorLLC's CLUES database.
        :rtype: str
        """
        # Search in Bluetooth known UUIDs
        alias = get_uuid_alias(self.__service_uuid)
        if alias is not None:
            return f"{alias} (0x{self.__service_uuid})"

        # Search in collaborative CLUES database
        alias = CluesDb.get_uuid_alias(self.__service_uuid)
        if alias is not None:
            if self.__service_uuid.type == UUID.TYPE_16:
                return f"{alias} (0x{self.__service_uuid})"
            else:
                return f"{alias} ({self.__service_uuid})"

        # Default name
        return str(self.__service_uuid)

    def payload(self):
        """Return service attribute's value.

        :return: Service attribute value
        :rtype: bytes
        """
        return self.__service_uuid.packed

    def add_characteristic(self, characteristic: Characteristic):
        """Add characteristic into this service.

        :param characteristic: Characteristic object to add into this service.
        :type  characteristic: Characteristic
        """
        # If characteristic's handle is 0, we define its handle before
        # adding it into our list of characteristics.
        if characteristic.handle == 0:
            characteristic.handle = self.end_handle + 1

        # Add this characteristic into our list
        self.__characteristics.append(characteristic)

        # Update our end handle
        self.__end_handle = max(characteristic.end_handle, self.__end_handle)

    def remove_characteristic(self, characteristic: Union[UUID, Type[Characteristic]]):
        """Remove a specific characteristic

        :param characteristic: Characteristic object to remove from service's characteristics.
        :type  characteristic: Characteristic, UUID
        """
        if isinstance(characteristic, UUID):
            # Look for characteristic and remove it if found
            for charac in self.__characteristics:
                if charac.uuid == characteristic:
                    self.__characteristics.remove(charac)
                    break
        elif isinstance(characteristic, Characteristic):
            # Look for characteristic object
            if characteristic in self.__characteristics:
                self.__characteristics.remove(characteristic)

        # Update characteristic handles
        char_handle = self.handle
        for charac in self.__characteristics:
            charac.handle = char_handle + 1
            char_handle = charac.end_handle

        # Update service end_handle value
        self.__end_handle = char_handle

    def characteristics(self) -> Iterator[Characteristic]:
        """Enumerate characteristics.

        :return: Iterator over this service's characteristics.
        """
        yield from self.__characteristics

    def get_characteristic(self, uuid: Union[str, UUID]) -> Optional[Characteristic]:
        """Get characteristic by UUID.

        .. deprecated:: 1.3.0
            Use :py:meth:`~whad.ble.profile.service.Service.char` instead of
            :py:meth:`.get_characteristic`.

        :param uuid: Searched characteristic's UUID
        :type uuid: UUID, str
        :return: Characteristic object if found, None otherwise
        :rtype: Characteristic, optional
        :raises InvalidUUIDException: Invalid UUID
        """
        return self.char(uuid)

    def char(self, uuid: Union[str, UUID]) -> Optional[Characteristic]:
        """Get characteristic by UUID.

        :param uuid: Searched characteristic's UUID
        :type uuid: UUID
        :return: Characteristic object if found, None otherwise
        :rtype: Characteristic, optional
        :raises InvalidUUIDException: Invalid UUID
        """
        # Convert characteristic's string UUID to an UUID object
        if isinstance(uuid, str):
            uuid = UUID(uuid)

        for charac in self.__characteristics:
            if charac.uuid == uuid:
                return charac
        return None

    def add_included_service(self, included_service: 'IncludeService'):
        """Add include service definition.

        :param included_service: Include service definition
        :type  included_service: IncludeService
        """
        # Set included service handle if not already set.
        if included_service.handle == 0:
            included_service.handle = self.end_handle + 1

        # Add included service into our list.
        self.__included_services.append(included_service)

        # Update end handle
        self.__end_handle = max(included_service.handle, self.__end_handle)

    def remove_include_service(self, included_service: Union[UUID, Type['IncludeService']]):
        """Remove a specific include service definition.

        :param included_service: Include service definition or its UUID
        :type  included_service: IncludeService, UUID
        """
        if isinstance(included_service, UUID):
            # Look for characteristic and remove it if found
            for inc_service in self.__included_services:
                if inc_service.uuid == included_service:
                    self.__included_services.remove(inc_service)
                    break
        elif isinstance(included_service, IncludeService):
            # Look for characteristic object
            if included_service in self.__included_services:
                self.__included_services.remove(included_service)

        # Update included services and characteristic handles
        char_handle = self.handle
        for inc_service in self.__included_services:
            inc_service.handle = char_handle + 1
            char_handle = inc_service.handle
        for characteristic in self.__characteristics:
            characteristic.handle = char_handle + 1
            char_handle = characteristic.end_handle

        # Update service end_handle value
        self.__end_handle = char_handle

    def included_services(self) -> Iterator['IncludeService']:
        """Enumerate included services
        """
        yield from self.__included_services

    @classmethod
    def _build(cls, instance: 'Service'):
        """Build a service based on the current object (template)."""
        assert instance.handle == 0

        # Create a basic service with same properties.
        service = cls(instance.uuid, instance.type_uuid, 0, 0)

        # Clone and add characteristics.
        for charac in instance.characteristics():
            service.add_characteristic(charac.build())
            if charac.alias is not None:
                setattr(service, charac.alias, charac)

        # Return our cloned service
        return service

    def build(self):
        """Build a new service based on current template class."""
        return self.__class__._build(self)

class PrimaryService(Service):
    """Primary service attribute.

    This attribute has a type UUID of 0x2800.
    """

    def __init__(self, uuid: UUID, handle: int = 0, end_handle: int = 0, **characteristics):
        """Initialize a primary service of UUID `uuid` and declare the requested characteristics.

        :param uuid: Service UUID
        :type  uuid: UUID
        :param handle: Service handle
        :type  handle: int, optional
        :param end_handle: Handle of service's last attribute
        :type  end_handle: int, optional
        :param characteristics: Additional characteristics's definitions to add into this service
        :type  characteristics: dict
        """
        super().__init__( UUID(0x2800), uuid, handle=handle, end_handle=end_handle, **characteristics)


    @classmethod
    def _build(cls, instance):
        """Clone service."""
        # Create a basic service with same properties.
        service = cls(instance.uuid, 0, 0)

        # Clone and add characteristics.
        for charac in instance.characteristics():
            charac_obj = charac.build()
            service.add_characteristic(charac_obj)
            if charac.alias:
                setattr(service, charac.alias, charac_obj)

        # Clone and add included services.
        for inc_service in instance.included_services():
            service.add_included_service(inc_service.build())

        # Return our cloned service
        return service

class ServiceEvent:
    """Default service event"""

class ServiceEventHandler(ABC):
    """Abstract class for service event handlers."""
    @abstractmethod
    def on_event(self, event: ServiceEvent):
        """Process a service event.

        :param event: Service event to process
        :type  event: ServiceEvent
        """

class StandardService(PrimaryService):
    """Standard service class.

    Service UUID shall be set as a property in every class inheriting from this
    service.
    """

    _uuid = None

    def __init__(self, handle: int = 0, end_handle: int = 0, **children):
        """Initialize a standard service.

        :param handle: Service start handle
        :type  handle: int, optional
        :param end_handle: Service end handle
        :type  end_handle: int, optional
        :param kwargs: Extra named parameters passed to the underlying :class:`PrimaryService`
        :type  kwargs: dict
        """
        self.__event_handlers = []
        super().__init__(self._uuid, handle, end_handle, **children)

    def add_event_handler(self, handler):
        """Add an event handler to this standard service."""
        if handler not in self.__event_handlers:
            self.__event_handlers.append(handler)

    def remove_event_handler(self, handler):
        """Remove event handler from this standard service."""
        if handler in self.__event_handlers:
            self.__event_handlers.remove(handler)

    def __threaded_event(self, handlers, event):
        """Send an event to handlers from a separate thread."""
        for handler in handlers:
            handler(event)

    def send_event(self, event):
        """Send event to registered event handlers."""
        # Get a copy of event handlers
        handlers = [handler for handler in self.__event_handlers]

        # Send event from another thread to avoid concurrency issues
        Thread(target=self.__threaded_event, args=(handlers, event)).start()

    @classmethod
    def _build(cls, instance):
        """Clone service."""
        # Create a basic service with same properties.
        service = cls(0, 0)

        # Return our cloned service
        return service

class SecondaryService(Service):
    """Secondary service attribute.

    This attribute has a type UUID of 0x2801.
    """

    def __init__(self, uuid, handle: int = 0):
        """Initialize a secondary service identified by UUID `uuid`,"""
        super().__init__( UUID(0x2801), uuid, handle=handle)

    @classmethod
    def _build(cls, instance):
        """Clone service."""
        # Create a basic service with same properties.
        service = cls(instance.uuid, 0)

        # Clone and add characteristics.
        for charac in instance.characteristics():
            service.add_characteristic(charac)
            if charac.alias:
                setattr(service, charac.alias, charac)

        # Return our cloned service
        return service

class IncludeService(Attribute):
    """IncludeService Attribute class

    This class stores the information related to an included service:

    - the included service UUID (16-bit or 128-bit UUID)
    - the start and end handles of the said included service
    """

    def __init__(self, uuid, handle=0, start_handle=0, end_handle=0):
        """Initialize an included service

        :param  uuid:       Included service UUID
        :type   uuid:       UUID
        :param  handle:     Included service start handle
        :type   handle:     int
        :param  end_handle: Included service end handle
        :type   end_handle: int
        """
        self.__service_uuid = uuid
        self.__start_handle = start_handle
        self.__end_handle = end_handle
        super().__init__(UUID(0x2802), handle=handle, value=self.payload())

    @property
    def end_handle(self):
        """Return this attribute end handle

        This attribute does not belong to a group so its end handle is the same
        than its handle.
        """
        return self.handle

    @end_handle.setter
    def end_handle(self, value) -> int:
        """End handle
        """
        self.__end_handle = value

    @property
    def uuid(self) -> UUID:
        """Return the attribute type UUID.
        """
        return self.type_uuid

    @property
    def service_uuid(self) -> UUID:
        """Return the included service UUID
        """
        return self.__service_uuid

    @property
    def service_start_handle(self) -> int:
        """Return the included service start handle
        """
        return self.__start_handle

    @service_start_handle.setter
    def service_start_handle(self, value: int):
        """Service start handle setter
        """
        self.__start_handle = value

    @property
    def service_end_handle(self) -> int:
        """Return the included service end handle
        """
        return self.__end_handle

    @service_end_handle.setter
    def service_end_handle(self, value: int):
        """End handle setter
        """
        self.__end_handle = value

    @property
    def name(self) -> str:
        """Generate the description of the included service definition attribute.
        """
        # Search in Bluetooth known database
        alias = get_uuid_alias(self.__service_uuid)
        if alias is not None:
            return f"Included service {alias} (0x{self.__service_uuid})"

        # Search in collaborative CLUES database
        alias = CluesDb.get_uuid_alias(self.__service_uuid)
        if alias is not None:
            if self.__service_uuid.type == UUID.TYPE_16:
                return f"Included service {alias} (0x{self.__service_uuid})"
            else:
                return f"Included service {alias} ({self.__service_uuid})"

        # Not found, default name is UUID
        return f"Included service {self.__service_uuid}"

    def payload(self) -> bytes:
        """Return service UUID as bytes
        """
        if self.__service_uuid.type == UUID.TYPE_16:
            return pack('<HH', self.__start_handle, self.__end_handle) + self.__service_uuid.packed
        # 128-bit UUID
        return pack('<HH', self.__start_handle, self.__end_handle)

    @classmethod
    def _build(cls, instance):
        """Clone service."""
        # Create a basic service with same properties.
        return cls(instance.uuid)




"""WHAD Protocol Discovery reset messages abstraction layer.
"""

from whad.protocol.whad_pb2 import Message
from whad.protocol.device_pb2 import DeviceResetQuery, DeviceReadyResp
from whad.hub.message import pb_bind, PbMessageWrapper
from whad.hub.discovery import Discovery

@pb_bind(Discovery, 'reset_query', 1)
class ResetQuery(PbMessageWrapper):
    """Device reset message class
    """
    
    def __init__(self, message: Message = None):
        """Define an empty message deriving from DeviceResetQuery.
        """
        super().__init__(message=message)
        self.message.discovery.reset_query.CopyFrom(DeviceResetQuery())

@pb_bind(Discovery, 'ready_resp', 1)
class DeviceReady(PbMessageWrapper):
    """Device ready message class
    """
    
    def __init__(self, message: Message = None):
        """Define an empty message deriving from DeviceReadyResp.
        """
        super().__init__(message=message)
        self.message.discovery.ready_resp.CopyFrom(DeviceReadyResp())


"""BLE Attribute
"""
from typing import Optional, List, Type, Tuple
from struct import pack, unpack

from whad.ble.exceptions import InvalidHandleValueException, InvalidUUIDException

# Default services UUID and associated names
SERVICES_UUID = {
    0x1800: "Generic Access",
    0x1801: "Generic Attribute",
    0x1802: "Immediate Alert",
    0x1803: "Link Loss",
    0x1804: "Tx Power",
    0x1805: "Current Time",
    0x1806: "Reference Time Update",
    0x1807: "Next DST Change",
    0x1808: "Glucose",
    0x1809: "Health Thermometer",
    0x180a: "Device Information",
    0x180d: "Heart Rate",
    0x180e: "Phone Alert Status",
    0x180f: "Battery",
    0x1810: "Blood Pressure",
    0x1811: "Alert Notification",
    0x1812: "Human Interface Device",
    0x1813: "Scan Parameters",
    0x1814: "Running Speed and Cadence",
    0x1815: "Automation IO",
    0x1816: "Cycling Speed and Cadence",
    0x1818: "Cycling Power",
    0x1819: "Location and Navigation",
    0x181a: "Environmental Sensing",
    0x181b: "Body Composition",
    0x181c: "User Data",
    0x181d: "Weight Scale",
    0x181e: "Bond Management",
    0x181f: "Continuous Glucose Monitoring",
    0x1820: "Internet Protocol Support",
    0x1821: "Indoor Positioning",
    0x1822: "Pulse Oximeter",
    0x1823: "HTTP Proxy",
    0x1824: "Transport Discovery",
    0x1825: "Object Transfer",
    0x1826: "Fitness Machine",
    0x1827: "Mesh Provisioning",
    0x1828: "Mesh Proxy",
    0x1829: "Reconnection Configuration",
    0x183a: "Insulin Delivery",
    0x183b: "Binary Sensor",
    0x183c: "Emergency Configuration",
    0x183d: "Authorization Control",
    0x183e: "Physical Activity Monitor",
    0x1843: "Audio Input Control",
    0x1844: "Volume Control",
    0x1845: "Volume Offset Control",
    0x1846: "Coordinated Set Identification",
    0x1847: "Device Time",
    0x1848: "Media Control",
    0x1849: "Generic Media Control",
    0x184a: "Constant Tone Extension",
    0x184b: "Telephone Bearer",
    0x184c: "Generic Telepehone Bearer",
    0x184d: "Microphone Control",
    0x184e: "Audio Stream Control",
    0x184f: "Broadcast Audio Scan",
    0x1850: "Published Audio Capabilities",
    0x1851: "Basic Audio Announcement",
    0x1852: "Broadcast Audio Announcement",
    0x1853: "Common Audio",
    0x1854: "Hearing Aid",
    0x1855: "TMAS",
    0x1856: "Public Broadcast Announcement"
}

# Default characteristics UUID and associated names
CHARACS_UUID = {
    0x2A00: "Device Name",
    0x2A01: "Appearance",
    0x2A02: "Peripheral Privacy Flag",
    0x2A03: "Reconnection Address",
    0x2A04: "Peripheral Preferred Connection Parameters",
    0x2A05: "Service Changed",
    0x2A06: "Alert Level",
    0x2A07: "Tx Power Level",
    0x2A08: "Date Time",
    0x2A09: "Day of Week",
    0x2A0A: "Day Date Time",
    0x2A0C: "Exact Time 256",
    0x2A0D: "DST Offset",
    0x2A0E: "Time Zone",
    0x2A0F: "Local Time Information",
    0x2A11: "Time with DST",
    0x2A12: "Time Accuracy",
    0x2A13: "Time Source",
    0x2A14: "Reference Time Information",
    0x2A16: "Time Update Control Point",
    0x2A17: "Time Update State",
    0x2A18: "Glucose Measurement",
    0x2A19: "Battery Level",
    0x2A1C: "Temperature Measurement",
    0x2A1D: "Temperature Type",
    0x2A1E: "Intermediate Temperature",
    0x2A21: "Measurement Interval",
    0x2A22: "Boot Keyboard Input Report",
    0x2A23: "System ID",
    0x2A24: "Model Number String",
    0x2A25: "Serial Number String",
    0x2A26: "Firmware Revision String",
    0x2A27: "Hardware Revision String",
    0x2A28: "Software Revision String",
    0x2A29: "Manufacturer Name String",
    0x2A2A: "IEEE 11073­20601 Regulatory Certification Data List",
    0x2A2B: "Current Time",
    0x2A2C: "Magnetic Declination",
    0x2A31: "Scan Refresh",
    0x2A32: "Boot Keyboard Output Report",
    0x2A33: "Boot Mouse Input Report",
    0x2A34: "Glucose Measurement Context",
    0x2A35: "Blood Pressure Measurement",
    0x2A36: "Intermediate Cuff Pressure",
    0x2A37: "Heart Rate Measurement",
    0x2A38: "Body Sensor Location",
    0x2A39: "Heart Rate Control Point",
    0x2A3F: "Alert Status",
    0x2A40: "Ringer Control Point",
    0x2A41: "Ringer Setting",
    0x2A42: "Alert Category ID Bit Mask",
    0x2A43: "Alert Category ID",
    0x2A44: "Alert Notification Control Point",
    0x2A45: "Unread Alert Status",
    0x2A46: "New Alert",
    0x2A47: "Supported New Alert Category",
    0x2A48: "Supported Unread Alert Category",
    0x2A49: "Blood Pressure Feature",
    0x2A4A: "HID Information",
    0x2A4B: "Report Map",
    0x2A4C: "HID Control Point",
    0x2A4D: "Report",
    0x2A4E: "Protocol Mode",
    0x2A4F: "Scan Interval Window",
    0x2A50: "PnP ID",
    0x2A51: "Glucose Feature",
    0x2A52: "Record Access Control Point",
    0x2A53: "RSC Measurement",
    0x2A54: "RSC Feature",
    0x2A55: "SC Control Point",
    0x2A5A: "Aggregate",
    0x2A5B: "CSC Measurement",
    0x2A5C: "CSC Feature",
    0x2A5D: "Sensor Location",
    0x2A5E: "PLX Spot­Check Measurement",
    0x2A5F: "PLX Continuous Measurement",
    0x2A60: "PLX Features",
    0x2A63: "Cycling Power Measurement",
    0x2A64: "Cycling Power Vector",
    0x2A65: "Cycling Power Feature",
    0x2A66: "Cycling Power Control Point",
    0x2A67: "Location and Speed",
    0x2A68: "Navigation",
    0x2A69: "Position Quality",
    0x2A6A: "LN Feature",
    0x2A6B: "LN Control Point",
    0x2A6C: "Elevation",
    0x2A6D: "Pressure",
    0x2A6E: "Temperature",
    0x2A6F: "Humidity",
    0x2A70: "True Wind Speed",
    0x2A71: "True Wind Direction",
    0x2A72: "Apparent Wind Speed",
    0x2A73: "Apparent Wind Direction",
    0x2A74: "Gust Factor",
    0x2A75: "Pollen Concentration",
    0x2A76: "UV Index",
    0x2A77: "Irradiance",
    0x2A78: "Rainfall",
    0x2A79: "Wind Chill",
    0x2A7A: "Heat Index",
    0x2A7B: "Dew Point",
    0x2A7D: "Descriptor Value Changed",
    0x2A7E: "Aerobic Heart Rate Lower Limit",
    0x2A7F: "Aerobic Threshold",
    0x2A80: "Age",
    0x2A81: "Anaerobic Heart Rate Lower Limit",
    0x2A82: "Anaerobic Heart Rate Upper Limit",
    0x2A83: "Anaerobic Threshold",
    0x2A84: "Aerobic Heart Rate Upper Limit",
    0x2A85: "Date of Birth",
    0x2A86: "Date of Threshold Assessment",
    0x2A87: "Email Address",
    0x2A88: "Fat Burn Heart Rate Lower Limit",
    0x2A89: "Fat Burn Heart Rate Upper Limit",
    0x2A8A: "First Name",
    0x2A8B: "Five Zone Heart Rate Limits",
    0x2A8C: "Gender",
    0x2A8D: "Heart Rate Max",
    0x2A8E: "Height",
    0x2A8F: "Hip Circumference",
    0x2A90: "Last Name",
    0x2A91: "Maximum Recommended Heart Rate",
    0x2A92: "Resting Heart Rate",
    0x2A93: "Sport Type for Aerobic and Anaerobic Thresholds",
    0x2A94: "Three Zone Heart Rate Limits",
    0x2A95: "Two Zone Heart Rate Limits",
    0x2A96: "VO2 Max",
    0x2A97: "Waist Circumference",
    0x2A98: "Weight",
    0x2A99: "Database Change Increment",
    0x2A9A: "User Index",
    0x2A9B: "Body Composition Feature",
    0x2A9C: "Body Composition Measurement",
    0x2A9D: "Weight Measurement",
    0x2A9E: "Weight Scale Feature",
    0x2A9F: "User Control Point",
    0x2AA0: "Magnetic Flux Density ­ 2D",
    0x2AA1: "Magnetic Flux Density ­ 3D",
    0x2AA2: "Language",
    0x2AA3: "Barometric Pressure Trend",
    0x2AA4: "Bond Management Control Point",
    0x2AA5: "Bond Management Feature",
    0x2AA6: "Central Address Resolution",
    0x2AA7: "CGM Measurement",
    0x2AA8: "CGM Feature",
    0x2AA9: "CGM Status",
    0x2AAA: "CGM Session Start Time",
    0x2AAB: "CGM Session Run Time",
    0x2AAC: "CGM Specific Ops Control Point",
    0x2AAD: "Indoor Positioning Configuration",
    0x2AAE: "Latitude",
    0x2AAF: "Longitude",
    0x2AB0: "Local North Coordinate",
    0x2AB1: "Local East Coordinate",
    0x2AB2: "Floor Number",
    0x2AB3: "Altitude",
    0x2AB4: "Uncertainty",
    0x2AB5: "Location Name",
    0x2AB6: "URI",
    0x2AB7: "HTTP Headers",
    0x2AB8: "HTTP Status Code",
    0x2AB9: "HTTP Entity Body",
    0x2ABA: "HTTP Control Point",
    0x2ABB: "HTTPS Security",
    0x2ABC: "TDS Control Point",
    0x2ABD: "OTS Feature",
    0x2ABE: "Object Name",
    0x2ABF: "Object Type",
    0x2AC0: "Object Size",
    0x2AC1: "Object First­Created",
    0x2AC2: "Object Last­Modified",
    0x2AC3: "Object ID",
    0x2AC4: "Object Properties",
    0x2AC5: "Object Action Control Point",
    0x2AC6: "Object List Control Point",
    0x2AC7: "Object List Filter",
    0x2AC8: "Object Changed",
    0x2AC9: "Resolvable Private Address Only",
    0x2ACC: "Fitness Machine Feature",
    0x2ACD: "Treadmill Data",
    0x2ACE: "Cross Trainer Data",
    0x2ACF: "Step Climber Data",
    0x2AD0: "Stair Climber Data",
    0x2AD1: "Rower Data",
    0x2AD2: "Indoor Bike Data",
    0x2AD3: "Training Status",
    0x2AD4: "Supported Speed Range",
    0x2AD5: "Supported Inclination Range",
    0x2AD6: "Supported Resistance Level Range",
    0x2AD7: "Supported Heart Rate Range",
    0x2AD8: "Supported Power Range",
    0x2AD9: "Fitness Machine Control Point",
    0x2ADA: "Fitness Machine Status",
    0x2ADB: "Mesh Provisioning Data In",
    0x2ADC: "Mesh Provisioning Data Out",
    0x2ADD: "Mesh Proxy Data In",
    0x2ADE: "Mesh Proxy Data Out",
    0x2AE0: "Average Current",
    0x2AE1: "Average Voltage",
    0x2AE2: "Boolean",
    0x2AE3: "Chromatic Distance from Planckian",
    0x2AE4: "Chromaticity Coordinates",
    0x2AE5: "Chromaticity in CCT and Duv Values",
    0x2AE6: "Chromaticity Tolerance",
    0x2AE7: "CIE 13.3­1995 Color Rendering Index",
    0x2AE8: "Coefficient",
    0x2AE9: "Correlated Color Temperature",
    0x2AEA: "Count 16",
    0x2AEB: "Count 24",
    0x2AEC: "Country Code",
    0x2AED: "Date UTC",
    0x2AEE: "Electric Current",
    0x2AEF: "Electric Current Range",
    0x2AF0: "Electric Current Specification",
    0x2AF1: "Electric Current Statistics",
    0x2AF2: "Energy",
    0x2AF3: "Energy in a Period of Day",
    0x2AF4: "Event Statistics",
    0x2AF5: "Fixed String 16",
    0x2AF6: "Fixed String 24",
    0x2AF7: "Fixed String 36",
    0x2AF8: "Fixed String 8",
    0x2AF9: "Generic Level",
    0x2AFA: "Global Trade Item Number",
    0x2AFB: "Illuminance",
    0x2AFC: "Luminous Efficacy",
    0x2AFD: "Luminous Energy",
    0x2AFE: "Luminous Exposure",
    0x2AFF: "Luminous Flux",
    0x2B00: "Luminous Flux Range",
    0x2B01: "Luminous Intensity",
    0x2B02: "Mass Flow",
    0x2B03: "Perceived Lightness",
    0x2B04: "Percentage 8",
    0x2B05: "Power",
    0x2B06: "Power Specification",
    0x2B07: "Relative Runtime in a Current Range",
    0x2B08: "Relative Runtime in a Generic Level Range",
    0x2B09: "Relative Value in a Voltage Range",
    0x2B0A: "Relative Value in an Illuminance Range",
    0x2B0B: "Relative Value in a Period of Day",
    0x2B0C: "Relative Value in a Temperature Range",
    0x2B0D: "Temperature 8",
    0x2B0E: "Temperature 8 in a Period of Day",
    0x2B0F: "Temperature 8 Statistics",
    0x2B10: "Temperature Range",
    0x2B11: "Temperature Statistics",
    0x2B12: "Time Decihour 8",
    0x2B13: "Time Exponential 8",
    0x2B14: "Time Hour 24",
    0x2B15: "Time Millisecond 24",
    0x2B16: "Time Second 16",
    0x2B17: "Time Second 8",
    0x2B18: "Voltage",
    0x2B19: "Voltage Specification",
    0x2B1A: "Voltage Statistics",
    0x2B1B: "Volume Flow",
    0x2B1C: "Chromaticity Coordinate",
    0x2B1D: "RC Feature",
    0x2B1E: "RC Settings",
    0x2B1F: "Reconnection Configuration Control Point",
    0x2B20: "IDD Status Changed",
    0x2B21: "IDD Status",
    0x2B22: "IDD Annunciation Status",
    0x2B23: "IDD Features",
    0x2B24: "IDD Status Reader Control Point",
    0x2B25: "IDD Command Control Point",
    0x2B26: "IDD Command Data",
    0x2B27: "IDD Record Access Control Point",
    0x2B28: "IDD History Data",
    0x2B29: "Client Supported Features",
    0x2B2A: "Database Hash",
    0x2B2B: "BSS Control Point",
    0x2B2C: "BSS Response",
    0x2B2D: "Emergency ID",
    0x2B2E: "Emergency Text",
    0x2B2F: "ACS Status",
    0x2B30: "ACS Data In",
    0x2B31: "ACS Data Out Notify",
    0x2B32: "ACS Data Out Indicate",
    0x2B33: "ACS Control Point",
    0x2B34: "Enhanced Blood Pressure Measurement",
    0x2B35: "Enhanced Intermediate Cuff Pressure",
    0x2B36: "Blood Pressure Record",
    0x2B37: "Registered User",
    0x2B38: "BR­EDR Handover Data",
    0x2B39: "Bluetooth SIG Data",
    0x2B3A: "Server Supported Features",
    0x2B3B: "Physical Activity Monitor Features",
    0x2B3C: "General Activity Instantaneous Data",
    0x2B3D: "General Activity Summary Data",
    0x2B3E: "CardioRespiratory Activity Instantaneous Data",
    0x2B3F: "CardioRespiratory Activity Summary Data",
    0x2B40: "Step Counter Activity Summary Data",
    0x2B41: "Sleep Activity Instantaneous Data",
    0x2B42: "Sleep Activity Summary Data",
    0x2B43: "Physical Activity Monitor Control Point",
    0x2B44: "Activity Current Session",
    0x2B45: "Physical Activity Session Descriptor",
    0x2B46: "Preferred Units",
    0x2B47: "High Resolution Height",
    0x2B48: "Middle Name",
    0x2B49: "Stride Length",
    0x2B4A: "Handedness",
    0x2B4B: "Device Wearing Position",
    0x2B4C: "Four Zone Heart Rate Limits",
    0x2B4D: "High Intensity Exercise Threshold",
    0x2B4E: "Activity Goal",
    0x2B4F: "Sedentary Interval Notification",
    0x2B50: "Caloric Intake",
    0x2B51: "TMAP Role",
    0x2B77: "Audio Input State",
    0x2B78: "Gain Settings Attribute",
    0x2B79: "Audio Input Type",
    0x2B7A: "Audio Input Status",
    0x2B7B: "Audio Input Control Point",
    0x2B7C: "Audio Input Description",
    0x2B7D: "Volume State",
    0x2B7E: "Volume Control Point",
    0x2B7F: "Volume Flags",
    0x2B80: "Volume Offset State",
    0x2B81: "Audio Location",
    0x2B82: "Volume Offset Control Point",
    0x2B83: "Audio Output Description",
    0x2B84: "Set Identity Resolving Key",
    0x2B85: "Coordinated Set Size",
    0x2B86: "Set Member Lock",
    0x2B87: "Set Member Rank",
    0x2B89: "Apparent Energy 32",
    0x2B8A: "Apparent Power",
    0x2B8C: "CO2 Concentration",
    0x2B8D: "Cosine of the Angle",
    0x2B8E: "Device Time Feature",
    0x2B8F: "Device Time Parameters",
    0x2B90: "Device Time",
    0x2B91: "Device Time Control Point",
    0x2B92: "Time Change Log Data",
    0x2B93: "Media Player Name",
    0x2B94: "Media Player Icon Object ID",
    0x2B95: "Media Player Icon URL",
    0x2B96: "Track Changed",
    0x2B97: "Track Title",
    0x2B98: "Track Duration",
    0x2B99: "Track Position",
    0x2B9A: "Playback Speed",
    0x2B9B: "Seeking Speed",
    0x2B9C: "Current Track Segments Object ID",
    0x2B9D: "Current Track Object ID",
    0x2B9E: "Next Track Object ID",
    0x2B9F: "Parent Group Object ID",
    0x2BA0: "Current Group Object ID",
    0x2BA1: "Playing Order",
    0x2BA2: "Playing Orders Supported",
    0x2BA3: "Media State",
    0x2BA4: "Media Control Point",
    0x2BA5: "Media Control Point Opcodes Supported",
    0x2BA6: "Search Results Object ID",
    0x2BA7: "Search Control Point",
    0x2BA8: "Energy 32",
    0x2BA9: "Media Player Icon Object Type",
    0x2BAA: "Track Segments Object Type",
    0x2BAB: "Track Object Type",
    0x2BAC: "Group Object Type",
    0x2BAD: "Constant Tone Extension Enable",
    0x2BAE: "Advertising Constant Tone Extension Minimum Length",
    0x2BAF: "Advertising Constant Tone Extension Minimum Transmit Count",
    0x2BB0: "Advertising Constant Tone Extension Transmit Duration",
    0x2BB1: "Advertising Constant Tone Extension Interval",
    0x2BB2: "Advertising Constant Tone Extension PHY",
    0x2BB3: "Bearer Provider Name",
    0x2BB4: "Bearer UCI",
    0x2BB5: "Bearer Technology",
    0x2BB6: "Bearer URI Schemes Supported List",
    0x2BB7: "Bearer Signal Strength",
    0x2BB8: "Bearer Signal Strength Reporting Interval",
    0x2BB9: "Bearer List Current Calls",
    0x2BBA: "Content Control ID",
    0x2BBB: "Status Flags",
    0x2BBC: "Incoming Call Target Bearer URI",
    0x2BBD: "Call State",
    0x2BBE: "Call Control Point",
    0x2BBF: "Call Control Point Optional Opcodes",
    0x2BC0: "Termination Reason",
    0x2BC1: "Incoming Call",
    0x2BC2: "Call Friendly Name",
    0x2BC3: "Mute",
    0x2BC4: "Sink ASE",
    0x2BC5: "Source ASE",
    0x2BC6: "ASE Control Point",
    0x2BC7: "Broadcast Audio Scan Control Point",
    0x2BC8: "Broadcast Receive State",
    0x2BC9: "Sink PAC",
    0x2BCA: "Sink Audio Locations",
    0x2BCB: "Source PAC",
    0x2BCC: "Source Audio Locations",
    0x2BCD: "Available Audio Contexts",
    0x2BCE: "Supported Audio Contexts",
    0x2BCF: "Ammonia Concentration",
    0x2BD0: "Carbon Monoxide Concentration",
    0x2BD1: "Methane Concentration",
    0x2BD2: "Nitrogen Dioxide Concentration",
    0x2BD3: "Non­Methane Volatile Organic Compounds Concentration",
    0x2BD4: "Ozone Concentration",
    0x2BD5: "Particulate Matter ­ PM1 Concentration",
    0x2BD6: "Particulate Matter ­ PM2.5 Concentration",
    0x2BD7: "Particulate Matter ­ PM10 Concentration",
    0x2BD8: "Sulfur Dioxide Concentration",
    0x2BD9: "Sulfur Hexafluoride Concentration",
    0x2BDA: "Hearing Aid Features",
    0x2BDB: "Hearing Aid Preset Control Point",
    0x2BDC: "Active Preset Index",
    0x2BDE: "Fixed String 64",
    0x2BDF: "High Temperature",
    0x2BE0: "High Voltage",
    0x2BE1: "Light Distribution",
    0x2BE2: "Light Output",
    0x2BE3: "Light Source Type",
    0x2BE4: "Noise",
    0x2BE5: "Relative Runtime in a Correlated Color Temperature Range",
    0x2BE6: "Time Second 32",
    0x2BE7: "VOC Concentration",
    0x2BE8: "Voltage Frequency"
}

# Descriptors UUID and associated names
DESCRIPTORS_UUID = {
    0x2900: "Characteristic Extended Properties",
    0x2901: "Characteristic User Description",
    0x2902: "Client Characteristic Configuration",
    0x2903: "Server Characteristic Configuration",
    0x2904: "Characteristic Presentation Format",
    0x2905: "Characteristic Aggregate Format"
}

class UUID:
    """UUID class borrowed from pyBT (c) Mike Ryan
    """
    TYPE_16 = 1
    TYPE_128 = 2

    uuid: str = ''
    packed: bytes = b''
    type: int = TYPE_16

    def __init__(self, uuid):
        if isinstance(uuid, UUID):
            self.uuid = uuid.uuid
            self.packed = uuid.packed
            self.type = uuid.type

        # integer
        elif isinstance(uuid, int):
            if 0 <= uuid <= 65536:
                self.uuid = f"{uuid:04X}"
                self.packed = pack('<H', uuid)
                self.type = UUID.TYPE_16
            elif 0 <= uuid <= 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF:
                self.uuid = f"{uuid:032X}"
                # modified solution from http://www.codegur.site/6877096/how-to-
                # pack-a-uuid-into-a-struct-in-python
                self.packed = pack('<QQ', uuid & 0xFFFFFFFFFFFFFFFF,
                                   (uuid >> 64) & 0xFFFFFFFFFFFFFFFF)
                self.type = UUID.TYPE_128

        elif len(uuid) == 4:
            self.uuid = uuid
            self.packed = bytes.fromhex(uuid)[::-1]
            self.type = UUID.TYPE_16
        elif len(uuid) == 36:
            temp = uuid.replace('-','')

            if len(temp) == 32:
                self.uuid = uuid
                self.packed = bytes.fromhex(temp)[::-1]
                self.type = UUID.TYPE_128
        elif len(uuid) == 32 and "-" not in uuid:
            self.uuid = b'-'.join((uuid[:8], uuid[8:12], uuid[12:16], uuid[16:20],
                                   uuid[20:])).decode('latin-1')
            self.packed = uuid.decode("hex")[::-1]
            self.type = UUID.TYPE_128
        # binary
        elif len(uuid) == 2:
            v = unpack('<H', uuid)[0]
            self.uuid = f"{v:04X}"
            self.packed = uuid
            self.type = UUID.TYPE_16
        elif len(uuid) == 16:
            r = uuid[::-1]
            self.uuid = "-".join(map(lambda x: x.hex(),
                                     (r[0:4], r[4:6], r[6:8],r[8:10], r[10:])))
            self.packed = uuid
            self.type = UUID.TYPE_128
        else:
            # Other format is not recognized as valid UUID
            raise InvalidUUIDException()

        if self.uuid is None:
            raise InvalidUUIDException(uuid)

    def __eq__(self, other):
        # TODO expand 16 bit UUIDs
        return self.packed == other.packed

    def __repr__(self):
        return self.uuid

    def to_bytes(self):
        """Convert UUID to bytes
        """
        return self.packed

    def value(self):
        """Retrieve UUID value (16-bit)
        """
        if self.type == UUID.TYPE_16:
            return unpack('<H', self.packed)[0]
        # Not a 16-bit UUID
        raise ValueError

    @classmethod
    def from_name(cls, name):
        """Get UUID from standard name.
        """
        return get_alias_uuid(name)

class Attribute:
    """GATT Attribute model
    """

    def __init__(self, uuid: UUID, handle: int = 0, value: bytes = b''):
        """Instantiate a GATT Attribute
        """
        self.__uuid = uuid
        self.__handle = handle
        self.__value = value

    def to_pdu(self):
        """Convert attribute to bytes (PDU)
        """
        return pack('<H', self.__handle) + self.__uuid.to_bytes() + self.payload()

    def payload(self):
        """Convert attribute to bytes
        """
        return self.__value

    @property
    def value(self) -> bytes:
        """Attribute value
        """
        return self.__value

    @value.setter
    def value(self, value: bytes):
        """Set attribute value
        """
        self.__value = value

    @property
    def handle(self) -> int:
        """Attribute handle
        """
        return self.__handle

    @handle.setter
    def handle(self, new_handle: int):
        """Set attribute handle
        """
        if isinstance(new_handle, int):
            self.__handle = new_handle

    @property
    def type_uuid(self) -> UUID:
        """Attribute type UUID
        """
        return self.__uuid

    def inspect(self, *classes: Type['Attribute']) -> List[Tuple[str, 'Attribute']]:
        """Inspect attribute class and return a list of attributes
        that match the expected class or classes."""
        attributes = []
        for name in dir(self):
            # Exclude special attributes
            if name.startswith('_'):
                continue

            # Retrieve attribute and check if its class matches
            attr = getattr(self, name)
            for match in classes:
                if isinstance(attr, match):
                    attributes.append((name, attr))

        # Return found attributes
        return attributes


def get_alias_uuid(alias: str):
    """Get 16-bit UUID from alias.
    """
    for uuid, name in SERVICES_UUID.items():
        if alias == name:
            return UUID(uuid)

    for uuid, name in CHARACS_UUID.items():
        if alias == name:
            return UUID(uuid)

    for uuid, name in DESCRIPTORS_UUID.items():
        if alias == name:
            return UUID(uuid)

    return None

def get_uuid_alias(uuid: UUID):
    """Get UUID alias for 16-bit UUID
    """
    if uuid.type == UUID.TYPE_16:
        uuid_val = int(uuid.uuid, 16)
        if uuid.type == UUID.TYPE_16:
            if uuid_val in SERVICES_UUID:
                return SERVICES_UUID[uuid_val]
            if uuid_val in CHARACS_UUID:
                return CHARACS_UUID[uuid_val]
            if uuid_val in DESCRIPTORS_UUID:
                return DESCRIPTORS_UUID[uuid_val]

    # Alias not found
    return None




"""Bluetooth Low Energy clone utility for WHAD

$ ble-clone --dump [bd address] profile.json
$ ble-clone -i hci0 profile.json
"""
from time import time
from json import loads, dumps
from json.decoder import JSONDecodeError

from hexdump import hexdump
from prompt_toolkit import print_formatted_text, HTML

from whad.cli.app import CommandLineApp, run_app
from whad.ble import Peripheral, GenericProfile, AdvDataFieldList
from whad.ble.connector import Central, Scanner
from whad.ble.stack.att.exceptions import AttError
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.cli.central.helpers import show_att_error
from whad.ble.profile.characteristic import CharacteristicProperties

def check_profile(profile):
    """Check profile validity.

    :param dict p: Profile
    :return bool: True if profile format is OK, False otherwise
    """
    # Make sure we have a 'devinfo' entry
    if 'devinfo' not in profile:
        return False

    # Make sure we have a valid adv_data entry
    if 'adv_data' not in profile['devinfo']:
        return False
    try:
        bytes.fromhex(profile["devinfo"]["adv_data"])
    except JSONDecodeError:
        return False
    except KeyError:
        return False
    except ValueError:
        return False

    # Make sure we have a valid scan_rsp entry (if provided)
    if 'scan_rsp' not in profile['devinfo']:
        return False
    if profile['devinfo']['scan_rsp'] is not None:
        try:
            bytes.fromhex(profile["devinfo"]["scan_rsp"])
        except JSONDecodeError:
            return False
        except KeyError:
            return False
        except ValueError:
            return False

    # Make sure we have a BD address
    if 'bd_addr' not in profile['devinfo']:
        return False

    # OK
    return True


class MonitoringProfile(GenericProfile):
    """Monitoring profile class used to display what is going with the emulated
    device.
    """


    def on_characteristic_read(self, service, characteristic, offset=0, length=0):
        """Characteristic read hook.

        This hook is called whenever a characteristic is about to be read by a GATT client.
        If this method returns a byte array, this byte array will be sent back to the
        GATT client. If this method returns None, then the read operation will return an
        error (not allowed to read characteristic value).
        

        :param BlePrimaryService service: Service owning the characteristic
        :param BleCharacteristic characteristic: Characteristic object
        :param int offset: Read offset (default: 0)
        :param int length: Max read length
        :return: Value to return to the GATT client
        """
        print_formatted_text(HTML((
            f"<ansigreen>Reading</ansigreen> characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(characteristic.value)

    def on_connect(self, conn_handle):
        print_formatted_text(HTML(
            "<ansired>New connection</ansired> handle:{conn_handle:d}"
        ))

    def on_disconnect(self, conn_handle):
        print_formatted_text(HTML(f"<ansired>Disconnection</ansired> handle:{conn_handle:d}"))

    def on_characteristic_written(self, service, characteristic, offset=0, value=b'',
                                  without_response=False):
        """Characteristic written hook

        This hook is called whenever a charactertistic has been written by a GATT
        client.
        """
        print_formatted_text(HTML((
            f"<ansimagenta>Wrote</ansimagenta> to characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(value)


    def on_characteristic_subscribed(self, service, characteristic, notification=False,
                                     indication=False):
        # Check if we have a hook to call
        print_formatted_text(HTML((
            f"<ansicyan>Subscribed</ansicyan> to characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))

    def on_characteristic_unsubscribed(self, service, characteristic):
        print_formatted_text(HTML((
            f"<ansicyan>Unsubscribed</ansicyan> to characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))

    def on_notification(self, service, characteristic, value):
        print_formatted_text(HTML((
            f"<ansicyan>Notification</ansicyan> from characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(value)

    def on_indication(self, service, characteristic, value):
        print_formatted_text(HTML((
            f"<ansicyan>Indication</ansicyan> from characteristic "
            f"<ansicyan>{characteristic.uuid}</ansicyan> of service "
            f"<ansicyan>{service.uuid}</ansicyan>"
        )))
        hexdump(value)

class BleCloneApp(CommandLineApp):
    """wble-clone CLI application class.
    """

    def __init__(self):
        super().__init__(
            description='WHAD Bluetooth Low Energy cloning utility',
            commands = False,
            interface = True
        )
        self.add_argument(
            '--dump',
            '-d',
            dest='dump',
            help='BD address of the device to dump'
        )
        self.add_argument('profile', metavar='PROFILE', help='Device profile file (JSON)')


    def dump_device(self, bdaddr, profile_path):
        """Create a dump of a device's profile
        """
        # Switch to Scanner mode, search our device
        scanner = Scanner(self.interface)
        scanner.start()

        start_time = time()
        device = None
        for dev in scanner.discover_devices():
            if dev.address.lower() == bdaddr.lower():
                if dev.got_scan_rsp or (time() - start_time) >= 10.0:
                    break

        # Device not found, stop
        if device is None:
            scanner.stop()
            return

        # Generate device advertising information
        if device.scan_rsp_records is not None:
            scan_rsp = device.scan_rsp_records.to_bytes().hex()
        else:
            scan_rsp = None

        device_metadata = {
            'adv_data': device.adv_records.to_bytes().hex(),
            'scan_rsp': scan_rsp,
            'bd_addr': str(device.address),
            'addr_type': device.address_type
        }
        scanner.stop()

        # Switch to central mode
        central = Central(self.interface)
        central.start()

        # Connect to target device
        device = central.connect(bdaddr)
        if device is None:
            self.error(f"Cannot connect to {bdaddr}, device does not respond.")
        else:
            # Perform profile discovery
            try:
                device.discover()
            except AttError as atterr:
                show_att_error(self, atterr)
                return
            except GattTimeoutException:
                self.error(
                    "GATT Timeout while discovering services and characteristics. Aborted."
                )
                return

            # Show discovered services and characteristics
            for service in device.services():
                print_formatted_text(HTML((
                    f"<ansicyan><b>Service {service.name}</b></ansicyan> "
                    f"(handle 0x{service.handle:04x} to 0x{service.end_handle:04x})"
                )))

                for charac in service.characteristics():
                    properties = charac.properties
                    rights = ''
                    if properties & CharacteristicProperties.READ != 0:
                        rights += '<ansigreen>R</ansigreen>'
                    if properties & CharacteristicProperties.WRITE != 0:
                        rights += '<ansired>W</ansired>'
                    if properties & CharacteristicProperties.INDICATE != 0:
                        rights += '<ansicyan>I</ansicyan>'
                    if properties & CharacteristicProperties.NOTIFY != 0:
                        rights += '<ansiblue>N</ansiblue>'

                    print_formatted_text(HTML((
                        f"  <b>{charac.name}</b> {rights}, handle 0x{charac.handle:x},"
                        f" value handle 0x{charac.value_handle:04x}"
                    )))

                    for desc in charac.descriptors():
                        print_formatted_text(HTML((
                           f"    Descriptor type {desc.type_uuid}, "
                           f"handle 0x{desc.handle:04x}"
                        )))
                print('')

            # Load GATT profile JSON data
            json_data = device.export_json()
            profile = loads(json_data)

            # Add specific device info (for emulating)
            profile['devinfo'] = device_metadata
            json_data = dumps(profile)
            try:
                print(f"Writing profile JSON data to {profile_path} ...")
                with open(profile_path, 'w', encoding="utf-8") as f:
                    f.write(json_data)
            except IOError:
                self.error(f"An error occurred when writing to {profile_path}")

            # Disconnect
            device.disconnect()

        # Terminate central
        central.stop()

    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        # Launch pre-run tasks
        self.pre_run()

        if self.args.interface is None:
            self.error('You must provide a WHAD interface with the --interface option.')
        else:
            # If dump is enabled
            if self.args.dump is not None:
                if self.args.profile is not None:
                    self.dump_device(self.args.dump, self.args.profile)
                else:
                    self.error('You must provide a target file to store the device profile.')
            elif self.args.profile is not None:
                # Emulate device
                with open(self.args.profile, 'r', encoding="utf-8") as f:
                    profile_json = f.read()
                    profile = loads(profile_json)

                # Check profile and emulate if everything is OK
                if check_profile(profile):
                    # Load device info
                    device_adv_data = bytes.fromhex(profile["devinfo"]["adv_data"])
                    device_scan_rsp = bytes.fromhex(profile["devinfo"]["scan_rsp"])

                    # Create a profile
                    device_profile = MonitoringProfile(from_json = profile_json)
                    periph = Peripheral(
                        self.interface,
                        profile=device_profile,
                        adv_data=AdvDataFieldList.from_bytes(device_adv_data),
                        scan_data=AdvDataFieldList.from_bytes(device_scan_rsp),
                        #bd_address=profile['devinfo']['bd_addr']
                    )
                    periph.start()
                    print((
                        f"Emulation of device {profile['devinfo']['bd_addr']} is active,"
                        f" press any key to stop..."))
                    try:
                        input()
                    except KeyboardInterrupt:
                        pass
                    print('\rStopping emulation ...')
                    periph.stop()
                else:
                    self.error('Bad JSON file format')

        # Launch post-run tasks
        self.post_run()

def ble_clone_main():
    """BLE clone application.
    """
    app = BleCloneApp()
    run_app(app)



"""
WHAD - Wireless HAcking Devices

This is the main WHAD client module.
"""
import logging

from whad.device import WhadDevice
from whad.exceptions import RequiredImplementation, UnsupportedDomain, \
    UnsupportedCapability, WhadDeviceNotReady, WhadDeviceNotFound, \
    WhadDeviceAccessDenied

# Force scapy to hide warnings
logging.getLogger("scapy.runtime").setLevel(logging.ERROR)


__all__ = [
    'UartDevice',
    'VirtualDevice',
    'WhadDevice',
    'RequiredImplementation',
    'UnsupportedDomain',
    'UnsupportedCapability',
    'WhadDeviceNotReady',
    'WhadDeviceNotFound',
    'WhadDeviceAccessDenied'
]



"""Bluetooth Low Energy ATT security mode helpers
"""
from whad.ble.stack.att.constants import BleAttSecurityMode

def get_att_security_mode_from_mode_and_level(mode, level):
    """
    Generate SecurityMode class instance based on a supplied mode and level.
    :param mode: Security mode
    :type mode: int
    :param level: Security level
    :type level: level
    :return: SecurityMode object corresponding to supplied values or False if DNE
    :rtype: SecurityMode || bool
    """
    if mode == 0 and level == 0:
        return BleAttSecurityMode.NO_ACCESS

    if mode == 1:
        if level == 1:
            return BleAttSecurityMode.OPEN
        if level == 2:
            return BleAttSecurityMode.ENCRYPTION_NO_AUTHENTICATION
        if level == 3:
            return BleAttSecurityMode.ENCRYPTION_WITH_AUTHENTICATION
        if level == 4:
            return BleAttSecurityMode.ENCRYPTION_WITH_SECURE_CONNECTIONS

    if mode == 2:
        if level == 1:
            return BleAttSecurityMode.DATA_SIGNING_NO_AUTHENTICATION
        if level == 2:
            return BleAttSecurityMode.DATA_SIGNING_WITH_AUTHENTICATION

    return False



"""GAP Heart Rate Service

This module provides an implementation of a standard Heart Rate service as defined in
the specification. This implementation supports the following features:
- Heart rate value update
- Energy expended update
- Skin contact depending on the configuration (default: no sensor supported)

The :class:`whad.ble.profile.services.hr.HeartRateService` class sends a specific
:class:`whad.ble.profile.services.hr.HeartRateService.UpdateEvent` event to registered
event handlers when used in a GATT client. This event allows a client application
to get updates as they are sent by the remote heart rate monitoring device.
"""
from struct import pack, unpack
from typing import Optional

from whad.ble.profile.attribute import UUID
from whad.ble.profile.device import PeripheralCharacteristic
from whad.ble.profile.service import StandardService, ServiceEvent
from whad.ble.profile.characteristic import Characteristic

class HeartRateService(StandardService):
    """Heart Rate service version 1.0 as defined in
    `specification <https://www.bluetooth.com/specifications/specs/html/?src=HRS_v1.0/out/en/index-en.html>`_.
    """

    class UpdateEvent(ServiceEvent):
        """Heart Rate service update event."""

        def __init__(self, rate: int, energy_expended: Optional[int], skin_contact: Optional[bool]):
            """Initialize an update event.

            :param rate: Heart rate value, in BPM
            :type  rate: int
            :param energy_expended: Energy expended, in kJ
            :type  energy_expended: int, optional
            :param skin_contact: Skin contact detected
            :type  skin_contact: bool, optional
            """
            self.__rate = rate
            self.__energy_expended = energy_expended
            self.__skin_contact = skin_contact

        @property
        def heart_rate(self) -> int:
            """Heart Rate value"""
            return self.__rate

        @property
        def energy_expended(self) -> Optional[int]:
            """Energy expended, if available."""
            return self.__energy_expended

        @property
        def skin_contact(self) -> Optional[bool]:
            """Skin contact detected, if sensor is present."""
            return self.__skin_contact

    # Service UUID
    _uuid = UUID(0x180d)

    # Flags
    FLAG_VALUE_FORMAT = 0x01
    FLAG_SENSOR_CONTACT = 0x02
    FLAG_SKIN_CONTACT = 0x04
    FLAG_ENERGY_EXP = 0x08
    FLAG_RR_INTERVAL = 0x10

    # Value format
    FORMAT_8BIT = 0
    FORMAT_16BIT = 1

    measurement = Characteristic(
        UUID(0x2a37),
        properties=Characteristic.NOTIFY,
        required=True,
        value=b''
    )

    body_sensor_location = Characteristic(
        UUID(0x2a38),
        properties=Characteristic.READ,
        value=b'Chest'
    )

    def __init__(self, handle: int = 0, end_handle: int = 0):
        """Initialize our heart rate service.

        :param handle: Service start handle
        :type  handle: int, optional
        :param end_handle: Service end handle
        :type  handle: int, optional
        :param contact: Enable sensor contact support
        :type  contact: bool
        """
        # Save information flags
        self.__contact = False
        self.__skin = False
        self.__heart_rate = 0
        self.__energy_expended = None

        # Initizalize this standard service
        super().__init__(handle=handle, end_handle=end_handle)

        # Subscribe for notifications if characteristic is of type `PeripheralCharacteristic`
        # (it means this service has been instantiated in a PeripheralProfile class and that
        # we act as a GATT client)
        if isinstance(self.measurement, PeripheralCharacteristic):
            self.measurement.subscribe(notification=True, callback=self.on_update)

    @property
    def location(self) -> str:
        return self.body_sensor_location.value.decode('utf-8')

    @property
    def rate(self) -> int:
        """Heart rate in BPM."""
        return self.__heart_rate

    @property
    def skin_contact(self) -> bool:
        """Skin contact status"""
        return self.__skin

    @property
    def energy_expended(self) -> Optional[int]:
        """Energy expended in kJ"""
        return self.__energy_expended

    @property
    def contact_sensor(self) -> bool:
        """Contact sensor present"""
        return self.__contact

    def enable_contact(self, contact: bool):
        """Enable/disable contact support."""
        self.__contact = contact

    def set_location(self, location: str):
        """Update body sensor location's characteristic value.

        :param location: Sensor location
        :type  location: str
        """
        self.body_sensor_location.value = location.encode('utf-8')

    def on_update(self, _: Characteristic, value: bytes, __: bool = False):
        """Process incoming notifications."""
        # We should at least get two bytes
        if len(value) < 2:
            raise ValueError()

        # Retrieve flags
        flags = value[0]
        offset = 1

        # Fetch contact and skin contact bits
        if (flags & self.FLAG_SENSOR_CONTACT) > 0:
            self.__contact = True
            if (flags & self.FLAG_SKIN_CONTACT) > 0:
                self.__skin = True
            else:
                self.__skin = False
        else:
            self.__contact = False
            self.__skin = False

        # Fetch heart rate value
        if (flags & self.FLAG_VALUE_FORMAT) > 0:
            # Check size
            if len(value) < 3:
                raise ValueError()
            # Parse
            self.__heart_rate = unpack('<H', value[1:3])[0]
            offset += 2
        else:
            self.__heart_rate = unpack('<B', value[1:2])[0]
            offset += 1

        # Fetch energy expended value if present
        if (flags & self.FLAG_ENERGY_EXP) > 0:
            # Check size
            if len(value) < (offset + 2):
                raise ValueError()
            # Parse value
            self.__energy_expended = unpack('<H', value[offset:offset+2])[0]
        else:
            self.__energy_expended = None

        # Send update event
        self.send_event(HeartRateService.UpdateEvent(
            self.__heart_rate,
            self.__energy_expended,
            self.__skin
        ))

    def update(self, rate: int, energy: Optional[int] = None, skin: Optional[bool] = None):
        """Update the measurement characteristic with the corresponding values.

        :param rate: Current heart rate in BPM
        :type  rate: int
        :param energy: Energy expended in kJ
        :type  energy: int, optional
        :param skin: Skin contact
        :type  skin: bool, optional
        """
        # Check heart rate value
        if rate < 0 or rate > 0xffff:
            raise ValueError()
        self.__heart_rate = rate

        # Build flags
        flags = 0
        if rate > 255:
            flags |= self.FLAG_VALUE_FORMAT

        # Check expended energy value and set flags
        if energy is not None:
            if energy < 0:
                raise ValueError()
            if energy > 0xffff:
                energy = 0xffff
            flags |= self.FLAG_ENERGY_EXP
            self.__energy_expended = energy

        # Check skin contact
        self.__contact = False
        if skin is not None and self.__contact:
            flags |= self.FLAG_SENSOR_CONTACT
            self.__contact = True
            if skin:
                flags |= self.FLAG_SKIN_CONTACT
                self.__skin = True
            else:
                self.__skin = False

        # Generate measure value
        value = b''
        if rate < 256:
            value += pack('<B', rate)
        else:
            value += pack('<H', rate)

        # Include expended energy if required
        if energy is not None:
            value += pack('<H', energy)

        # Update our measurement characteristic
        self.measurement.value = bytes([flags]) + value




"""Phy Packet translator
"""
from whad.scapy.layers.phy import Phy_Packet_Hdr, Phy_Packet
from whad.protocol.whad_pb2 import Message
from whad.hub import ProtocolHub
from whad.hub.phy import PacketReceived, RawPacketReceived, generate_phy_metadata
import logging

logger = logging.getLogger(__name__)

class PhyMessageTranslator(object):
    """Phy Whad message translator.

    This translator is used to provide the format of a specific scapy packet
    as well as standard methods to convert WHAD Phy messages into scapy packets
    (if it makes sense) and scapy packets into WHAD Phy messages.
    """

    def __init__(self, protocol_hub: ProtocolHub):
        # Address cache
        self.address = None
        self.pattern_cropped_bytes = 0
        self.pattern = None

        self.physical_layer = None

        self.__hub = protocol_hub

    def format(self, packet):
        """
        Converts a scapy packet with its metadata to a tuple containing a scapy packet with
        the appropriate header and the timestamp in microseconds.
        """
        return Phy_Packet_Hdr(rssi=packet.metadata.rssi, frequency=packet.metadata.frequency) / packet, None


    def from_message(self, message):
        if isinstance(message, RawPacketReceived):
            bytes_packet = bytes(message.packet)
            if self.pattern_cropped_bytes > 0:
                bytes_packet = self.pattern[:self.pattern_cropped_bytes] + bytes_packet

            if self.physical_layer is not None and self.physical_layer.decoding is not None:
                bytes_packet = self.physical_layer.decoding(bytes_packet, self.physical_layer.configuration)
            if bytes_packet is not None and len(bytes_packet) > 0:
                if self.physical_layer is not None and self.physical_layer.scapy_layer is not None:
                    packet = self.physical_layer.scapy_layer(bytes_packet)
                else:
                    packet = Phy_Packet(bytes_packet)
                    packet.metadata = generate_phy_metadata(message)

                return packet

        elif isinstance(message, PacketReceived):
            bytes_packet = bytes(message.packet)
            if self.pattern_cropped_bytes > 0:

                bytes_packet = self.pattern[:self.pattern_cropped_bytes] + bytes_packet
            if self.physical_layer is not None and self.physical_layer.decoding is not None:
                bytes_packet = self.physical_layer.decoding(bytes_packet, self.physical_layer.configuration)
            if bytes_packet is not None and len(bytes_packet) > 0:
                if self.physical_layer is not None and self.physical_layer.scapy_layer is not None:
                    packet = self.physical_layer.scapy_layer(bytes_packet)
                else:
                    packet = Phy_Packet(bytes_packet)
                    packet.metadata = generate_phy_metadata(message)
                return packet


    def from_packet(self, packet):
        if Phy_Packet in packet or isinstance(packet, bytes):
            msg = self.__hub.phy.create_send_packet(bytes(packet))
        elif isinstance(packet, list) and len(packet) % 2 == 0:
            msg = self.__hub.phy.create_send_raw_packet(packet)
        else:
            msg = None
        return msg



"""
WHAD Enhanced ShockBurst sniffing configuration

This module defines the ESB configuration class for
packet sniffing.
"""
from dataclasses import dataclass

@dataclass
class SnifferConfiguration:
    """
    Configuration for sniffing a Enhanced ShockBurst communication.

    :param channel: select the channel to sniff (c)
    :param address: provide address to sniff (f)
    :param scanning: enable scanning mode (s)
    :param acknowledgements: enable acknowledgements sniffing (a)
    """
    channel : int = 0
    address : str = "FF:FF:FF:FF:FF"
    scanning : bool = False
    acknowledgements : bool = False



"""
FindInformation procedure

The ReadByType procedure is initiated by a ReadByType request, and
based on the required attribute start/end and type returns a
ReadByType response or an error.
"""
from typing import List
from struct import pack

from scapy.packet import Packet
from scapy.layers.bluetooth import (
    ATT_Error_Response, ATT_Find_Information_Request, ATT_Find_Information_Response,
    ATT_Find_By_Type_Value_Request, ATT_Find_By_Type_Value_Response
)

from whad.ble.profile.attribute import UUID
from whad.ble.stack.att.constants import BleAttErrorCode, BleAttOpcode

from .attribute import find_attr_by_range, find_attr_by_type
from .procedure import BleServerProcedure, BleClientProcedure, UnexpectedProcError

class ServerFindInformationProcedure(BleServerProcedure):
    """ATT FindInformation procedure."""

    # ATT Operation code
    OPCODE = BleAttOpcode.FIND_INFO_REQUEST

    def __init__(self, attributes: list, mtu: int):
        """Initialize our ServerFindInformation procedure."""
        super().__init__(attributes, mtu)

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure should be triggered."""
        return ATT_Find_Information_Request in request

    def process_request(self, request: Packet) -> list[Packet]:
        """React only on a ATT_Find_Information_Request."""
        if ATT_Find_Information_Request not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        # Extract request
        request = request[ATT_Find_Information_Request]

        # List attributes with handles between start and end handles
        attrs = find_attr_by_range(self.attributes, start_handle=request.start, end_handle=request.end)
        if len(attrs) == 0:
            self.set_state(self.states.DONE)
            return self.att_error_response(request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

        # Build response
        resp = b""
        attr_uuid_type = None
        for attr in attrs:
            if attr_uuid_type is None:
                attr_uuid_type = attr.uuid.type
            if attr_uuid_type == attr.uuid.type:
                attr_data = pack("<H", attr.handle) + attr.uuid.packed
                if self.mtu - 2 - len(resp) >= len(attr_data):
                    resp += attr_data
                else:
                    break
            else:
                break

        # Send response
        self.set_state(self.states.DONE)
        return [
            ATT_Find_Information_Response(
                format=1 if attr_uuid_type == UUID.TYPE_16 else 2,
                handles=resp
            )
        ]

class ClientFindInformationProcedure(BleClientProcedure):
    """GATT Client FindInformation procedure."""

    # ATT Operation code
    OPCODE = BleAttOpcode.FIND_INFO_RESPONSE

    def __init__(self, start_handle: int, end_handle: int):
        """Initialize a FindInformation procedure.

        :param start_handle: Start handle
        :type start_handle: int
        :param end_handle: End handle
        :type end_handle: int
        """
        # Initialize our parent Procedure
        super().__init__([
            ATT_Find_Information_Request(
                start=start_handle,
                end=end_handle
            )
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming packets, we expect to receive ATT_Error_Response or ATT_Find_Information_Response
        packets.

        :param packet: Received packet
        :type packet: scapy.packet.Packet
        :return: A list of packets to send in response to the received packets
        :rtype: list
        """
        # Did we receive an ATT error ?
        if ATT_Error_Response in request:
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
            return []

        # Or did we receive an ATT_Find_Information_Response ?
        if ATT_Find_Information_Response in request:
            # Extract response
            response = request[ATT_Find_Information_Response]

            # Parse response content
            if response.format not in (1, 2):
                # Error, unknown format value
                self.set_result(None)
                self.set_state(self.states.ERROR)
                return []

            # Build list
            self.set_result(response.handles)
            self.set_state(self.states.DONE)
            return []
        else:
            self.set_result(None)
            self.set_state(self.states.ERROR)
            return []

class ServerFindByTypeValueProcedure(BleServerProcedure):
    """ATT FindInformation procedure."""

    # ATT Operation code
    OPCODE = BleAttOpcode.FIND_BY_TYPE_VALUE_REQUEST

    def __init__(self, attributes: list, mtu: int):
        """Initialize our ServerFindByTypeValue procedure."""
        super().__init__(attributes, mtu)

    @classmethod
    def trigger(cls, request) -> bool:
        """Determine if the procedure should be triggered."""
        return ATT_Find_By_Type_Value_Request in request

    def process_request(self, request: Packet) -> list[Packet]:
        """React only on a ATT_Find_Information_Request."""
        if ATT_Find_By_Type_Value_Request not in request:
            self.set_state(self.states.ERROR)
            raise UnexpectedProcError()

        # Extract request
        request = request[ATT_Find_By_Type_Value_Request]

        # List attributes with handles between start and end handles
        attrs = find_attr_by_type(self.attributes, UUID(request.uuid), start_handle=request.start, end_handle=request.end)
        if len(attrs) == 0:
            self.set_state(self.states.DONE)
            return self.att_error_response(request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

        # Build response
        resp = b""
        attr_uuid_type = None
        for attr in attrs:
            # If attribute value does not match, skip attribute.
            if attr.value != request.data:
                continue
            if attr_uuid_type is None:
                attr_uuid_type = attr.uuid.type
            if attr_uuid_type == attr.uuid.type:
                attr_data = pack("<HH", attr.handle, attr.end_handle)
                if self.mtu - 2 - len(resp) >= len(attr_data):
                    resp += attr_data
                else:
                    break
            else:
                break

        # If we did not find any service, return an error
        if resp == b'':
            self.set_state(self.states.DONE)
            return self.att_error_response(request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND)

        # Send response
        self.set_state(self.states.DONE)
        return [
            ATT_Find_By_Type_Value_Response(
                handles=resp
            )
        ]

class ClientFindByTypeValueProcedure(BleClientProcedure):
    """GATT Client FindByTypeValue procedure."""

    def __init__(self, start_handle: int, end_handle: int, attr_type: UUID, attr_value: bytes):
        """Initialize a FindByTypeValue procedure.

        :param start_handle: Start handle
        :type start_handle: int
        :param end_handle: End handle
        :type end_handle: int
        :param attr_type: Attribute Type UUID
        :type attr_type: UUID
        :param attr_value: Attribute value
        :type attr_value: bytes
        """
        # Initialize our parent Procedure
        super().__init__([
            ATT_Find_By_Type_Value_Request(
                start=start_handle,
                end=end_handle,
                uuid=attr_type.value(),
                data=attr_value
            )
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming packets, we expect to receive ATT_Error_Response or ATT_Find_Information_Response
        packets.

        :param packet: Received packet
        :type packet: scapy.packet.Packet
        :return: A list of packets to send in response to the received packets
        :rtype: list
        """
        # Did we receive an ATT error ?
        if ATT_Error_Response in request:
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
            return []

        # Or did we receive an ATT_Find_Information_Response ?
        if ATT_Find_By_Type_Value_Response in request:
            # Extract response
            response = request[ATT_Find_By_Type_Value_Response]
            self.set_result(response)
            self.set_state(self.states.DONE)
            return []
        else:
            self.set_result(None)
            self.set_state(self.states.ERROR)
            return []




from re import I
from typing import List

from scapy.packet import Packet
from scapy.layers.bluetooth import (
    ATT_Error_Response, ATT_Handle_Value_Indication, ATT_Hdr, ATT_Write_Request, ATT_Write_Response, ATT_Handle_Value_Notification,
)

from whad.ble.stack.att.constants import BleAttOpcode
from .procedure import BleClientProcedure

class ClientNotificationCheckProcedure(BleClientProcedure):
    """GATT client notification subscription procedure."""

    OPCODE = BleAttOpcode.HANDLE_VALUE_NOTIFICATION

    def __init__(self, handle: int):
        """Initialize a GATT client notfication check."""
        super().__init__([
            ATT_Write_Request(
                gatt_handle=handle,
                data=bytes([0x01, 0x00])
            )
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming PDUs."""
        request.show()
        # Do we got an error ? Force state to ERROR and save
        # the error details. Procedure is considered terminated.
        if ATT_Error_Response in request:
            print(f"Error received while in state {self.get_state()}")
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
            return []

        # We are expecting a write response following our Initial
        # write into the target attribute (CCC descriptor)
        if self.get_state() == self.states.INITIAL:
            if ATT_Hdr in request and request[ATT_Hdr].opcode == 0x13:
                print(f"Descriptor successfully modified, wait for notification ...")
                # Write response received, update state
                self.set_state(self.states.SUB_DONE)
                return []

        # Once subed to a characteristic, we are expecting a notification
        # to complete this procedure.
        if self.get_state() == self.states.SUB_DONE:
            # If we received a notification, everything is fine
            # and procedure is done.
            if ATT_Handle_Value_Notification in request:
                self.set_result(request[ATT_Handle_Value_Notification])
                self.set_state(self.states.DONE)
                return []

        # Default handler
        return []

class ClientIndicationCheckProcedure(BleClientProcedure):
    """GATT client indication subscription procedure."""

    OPCODE = BleAttOpcode.HANDLE_VALUE_INDICATION

    def __init__(self, handle: int):
        """Initialize a GATT client indication check."""
        super().__init__([
            ATT_Write_Request(
                gatt_handle=handle,
                data=bytes([0x02, 0x00])
            )
        ])

    def process_request(self, request: Packet) -> List[Packet]:
        """Process incoming PDUs."""
        request.show()
        # Do we got an error ? Force state to ERROR and save
        # the error details. Procedure is considered terminated.
        if ATT_Error_Response in request:
            print(f"Error received while in state {self.get_state()}")
            self.set_result(request[ATT_Error_Response])
            self.set_state(self.states.ERROR)
            return []

        # We are expecting a write response following our Initial
        # write into the target attribute (CCC descriptor)
        if self.get_state() == self.states.INITIAL:
            if ATT_Hdr in request and request[ATT_Hdr].opcode == 0x13:
                print(f"Descriptor successfully modified, wait for indication ...")
                # Write response received, update state
                self.set_state(self.states.SUB_DONE)
                return []

        # Once subed to a characteristic, we are expecting a indication
        # to complete this procedure.
        if self.get_state() == self.states.SUB_DONE:
            # If we received a notification, everything is fine
            # and procedure is done.
            if ATT_Handle_Value_Indication in request:
                print("SUCCESS")
                self.set_result(request[ATT_Handle_Value_Indication])
                self.set_state(self.states.DONE)
                return [ATT_Hdr(opcode=BleAttOpcode.HANDLE_VALUE_CONFIRMATION)]

        # Default handler
        return []




"""WHAD CLI Interactive shell module
"""
import re
import shlex
from prompt_toolkit import PromptSession, HTML, print_formatted_text
from prompt_toolkit.completion import NestedCompleter, DynamicCompleter
from whad.cli.ui import error, warning, success

# decorator for category
#pylint: disable-next=invalid-name
class category:
    """Shell command handler decorator.

    This decorator adds a category to the command handler (do_<something>)
    that will be used to group commands in help menu.
    """

    def __init__(self, cat: str):
        self.category = cat

    def __call__(self, handler):
        handler.category = self.category
        return handler


class InteractiveShell:

    """Interactive shell for WHAD CLI applications.

    This class introduce a behavior similar to Python's `cmd` module:
     - each command is implemented in a `do_<command>` method
     - for each command, a `complete_<command>` method can be implemented to
       provide contextual parameter completion
     - exit/quit commands are automatically handled

    Commands are automatically documented through their respective docstrings,
    as shown below:

    ```
    do_something(self, args):
        '''Do something

        <ansicyan><b>something</b> <i>[ some param ]</i></ansicyan>

        This command does something.
        '''
        pass
    ```

    Docstrings can use prompt-toolkit HTML to add colors and effects.
    """

    def __init__(self, default_prompt='> '):
        self.__commands = {}
        self.__commands_ac = {}
        self.__categories = {}
        self.__prompt = default_prompt
        self.__quit = False
        self.__env = {}
        self.__session = None

        # Search specific commands named 'do_$command$'
        for member_name in dir(self):
            member = getattr(self, member_name)
            if callable(member) and member_name.startswith('do_'):
                # Found a do_* method
                command = member_name[3:]
                if len(command) >= 1:
                    # Associate command with method
                    self.__commands[command] = member

                    # Add category if defined
                    if hasattr(member, 'category'):
                        member_category = getattr(member, 'category')
                        if member_category not in self.__categories:
                            self.__categories[member_category] = []
                        self.__categories[member_category].append(command)

                    # Look for autocomplete method
                    if hasattr(self, f"complete_{command}"):
                        command_ac = getattr(self, f"complete_{command}")
                        if callable(command_ac):
                            # Save autocomplete method
                            self.__commands_ac[command] = command_ac
                    else:
                        # No autocomplete found
                        self.__commands_ac[command] = None


    def set_prompt(self, prompt, force_update=False):
        """Set interactive shell prompt.

        :param str prompt: Prompt
        """
        self.__prompt = prompt
        if force_update and self.__session is not None:
            # We modify the `_message` attribute of our session
            # from the outside
            self.__session.message = self.__prompt

    def update_autocomplete(self):
        """Update nested autocompleter.

        This method will ask each command handler to autocomplete based on
        current state.
        """
        commands_autocomplete = {}
        for name, cmd in self.__commands_ac.items():
            if cmd is not None:
                nested_autocomplete = cmd()
                commands_autocomplete[name] = nested_autocomplete
            else:
                commands_autocomplete[name] = {}
        return NestedCompleter.from_nested_dict(commands_autocomplete)

    def autocomplete_env(self, pattern=None):
        """Return a list of current environment variables.

        :param str pattern: regular expression to filter variables
        """
        completions = {}
        for name, value in self.__env.items():
            if pattern is not None:
                if re.match(pattern, value):
                    completions['$'+name] = {}
            else:
                completions['$'+name] = {}
        return completions

    def process(self, input_line):
        """Process input commands.
        """
        try:
            # Dispatch commands
            tokens = shlex.split(input_line)
            if len(tokens) >= 1:
                command = tokens[0]
                if command in self.__commands:
                    resolved_args = [self.resolve(arg) for arg in tokens[1:]]
                    try:
                        # Command is supported, follow to method
                        return self.__commands[command](resolved_args)
                    except KeyboardInterrupt:
                        print("\rInterrupted by user.")
        except ValueError as e:
            print(e)
            self.warning('An error occurred while processing your command.')

    def run(self):
        """Run the interactive shell.
        """
        try:
            self.__session = PromptSession(completer=DynamicCompleter(self.update_autocomplete))
            while not self.__quit:
                input_line = self.__session.prompt(self.__prompt)
                res = self.process(input_line)
                if res:
                    break
        except KeyboardInterrupt:
            # Call do_quit() to terminate
            self.do_quit([])
        except EOFError:
            # Call do_quit() to terminate
            self.do_quit([])

    def run_script(self, script_path):
        """Run a script

        :param str script_path: path to a script file
        """
        try:
            # run script
            with open(script_path, 'r', encoding="utf-8") as script:
                commands = script.readlines()
                for cmd in commands:
                    res = self.process(cmd)
                    if res:
                        break
        except IOError:
            self.error(f"Cannot access script file \"{script_path}\"")

    def do_quit(self, args):
        """This method must be overriden to handle tool termination.
        """

    def stop(self):
        """Stop the interactive shell.
        """
        self.__quit = True

    def complete_help(self):
        """Auto-complete help command.
        """
        completions = {}
        for command in self.__commands:
            completions[command] = {}
        return completions

    def do_help(self, args):
        """show this help screen

        <ansicyan><b>help</b> <i>[ command ]</i></ansicyan>

        Shows help about the given command, if specified.
        """
        if len(args) >= 1:
            # Target command specified
            command = args[0]

            if command in self.__commands:
                # Get command docstring
                handler = self.__commands[command]
                if handler is not None and hasattr(handler, '__doc__'):
                    # Read docstring
                    docstr = getattr(handler,'__doc__')
                    desc = '\n'.join([l.lstrip() for l in docstr.splitlines()[1:]])
                    print_formatted_text(HTML(desc.strip()))
                    print('')
                else:
                    self.warning(f"command <u>{command}</u> is not documented")
            else:
                self.error(f"command <u>{command}</u> does not exist.")
        else:
            max_cmd_size = max([len(cmd) for cmd in list(self.__commands.keys())])

            # List available commands, categories first and then uncategorized.
            categorized_commands = []
            for name, cat_obj in self.__categories.items():
                commands = []
                for command in cat_obj:
                    # Get command docstring
                    handler = self.__commands[command]
                    if handler is not None and hasattr(handler, '__doc__'):
                        # Read docstring
                        docstr = getattr(handler,'__doc__')
                        short_desc = docstr.splitlines()[0].lstrip()
                    else:
                        short_desc = ''
                    commands.append((command, short_desc))
                    categorized_commands.append(command)

                # Show commands
                print_formatted_text(HTML(f"<ansimagenta><b>{name}:</b></ansimagenta>"))
                #max_cmd_size = max([len(cmd) for cmd,doc in commands])
                cmd_fmt = "  <ansicyan>{0:<%d}</ansicyan>\t\t{1}" % max_cmd_size
                for cmd, doc in commands:
                    print_formatted_text(HTML(cmd_fmt.format(cmd, doc)))
                print('')

            commands = []
            for name, cmd in self.__commands.items():
                if name not in categorized_commands:
                    # Get command docstring
                    if cmd is not None and hasattr(cmd, '__doc__'):
                        # Read docstring
                        docstr = getattr(cmd,'__doc__')
                        short_desc = docstr.splitlines()[0].lstrip()
                    else:
                        short_desc = ''
                    commands.append((name, short_desc))

            # Show commands
            print_formatted_text(HTML('<ansimagenta><b>Generic commands:</b></ansimagenta>'))
            #max_cmd_size = max([len(cmd) for cmd,doc in commands])
            cmd_fmt = "  <ansicyan>{0:<%d}</ansicyan>\t\t{1}" % max_cmd_size
            for cmd, doc in commands:
                print_formatted_text(HTML(cmd_fmt.format(cmd, doc)))


    def resolve(self, arg):
        """Resolve a shell parameter.

        This method checks if parameter is an environment variable and replace
        it with its value, or return the original parameter.

        :param str arg: shell parameter to resolve
        :return str: resolved parameter
        """
        if arg.startswith('$'):
            if arg[1:] in self.__env:
                return self.__env[arg[1:]]
            return arg

        if arg.startswith('\\$'):
            return arg[1:]

        return arg

    def do_set(self, args):
        """set environment variable

        <ansicyan><b>set</b> <i>ENV_VAR</i> <i>value</i></ansicyan>

        Set <i>ENV_VAR</i> environment variable to <i>value</i> and keep it in
        memory for the current session. This variable can be recalled with the
        following notation: <i>$ENV_VAR</i>.

        Variable's name cannot include digits, and only '_' is allowed as special
        char.
        """
        if len(args)>=2:
            varname = args[0]
            varval = args[1]

            # make sure name matches requirements
            if re.match('^[a-zA-Z_]+$', varname):
                # Store variable in environment
                self.__env[varname] = self.resolve(varval)
            else:
                self.error('Variable name contains an invalid character')
        else:
            self.error('Missing argument (see help)')

    def do_env(self, _):
        """show environment variables

        <ansicyan><b>env</b></ansicyan>

        List environment variables.
        """
        for name, value in self.__env.items():
            print_formatted_text(HTML(f"<ansicyan>{name}</ansicyan>={value}"))

    def do_unset(self, args):
        """remove environment variable

        <ansicyan><b>unset</b> <i>VAR_NAME</i></ansicyan>

        Remove <i>VAR_NAME</i> from environment.
        """
        if len(args)>=1:
            varname = args[0]
            if varname in self.__env:
                del self.__env[varname]

    def do_wait(self, args):
        """show a message and wait for the user to press a key

        <ansicyan><b>wait</b> <i>[MESSAGE]</i></ansicyan>

        Display <i>MESSAGE</i> in the standard output and wait for
        the user to press ENTER.

        If message is not provided, use a default message.
        """
        if len(args)>=1:
            message = args[0]
        else:
            message = "Press ENTER to continue..."

        # Show message and wait for a keypress
        try:
            print_formatted_text(HTML(f"<i>{message}</i>"))
            input()
        except EOFError:
            self.warning("Cannot wait for keyboard input, is your stdin piped to another app ?")

    def do_echo(self, args):
        """Display a text in the standard output.

        <ansicyan><b>echo</b> <i>TEXT</i></ansicyan>

        Print <i>TEXT</i> to the standard output.
        """
        # If at least one argument is provided
        if len(args)>=1:
            # Concatenate text arguments if multiple are provided
            print("".join(args))


    def success(self, message):
        """Display a success message in green (if color is enabled)
        """
        success(message)

    def warning(self, message):
        """Display a warning message in orange (if color is enabled)
        """
        warning(message)

    def error(self, message):
        """Display an error message in red (if color is enabled)
        """
        error(message)

if __name__ == '__main__':
    class MyShell(InteractiveShell):
        """Interactive shell wrapper for testing.
        """
        def __init__(self):
            """Constructor.
            """
            super().__init__('> ')

        def complete_test(self):
            """Test auto-complete.
            """
            return {'this':None, 'that':None}

        def do_test(self, args):
            print(args)

    ishell = MyShell()
    ishell.run()



"""
This module provides the :class:`whad.esb.connector.ptx.PTX` connector class
that implements Nordic Semiconductor's *Enhanced ShockBurst* *Primary transmitter*
role.

The *primary transmitter* (*PTX*) role consists in sending packets to another
*Enhanced ShockBurst* device in *primary receiver* (*PRX*) role and determine
if they have correctly been received. Correct packet reception is checked through
*acknowledgements* packets sent by the device in *PRX* mode, if required by
the device in *PTX* mode.

The :class:`whad.esb.connector.ptx.PTX` connector embeds a custom stack to
provide this behaviour and exposes some methods to send data. 
"""
from scapy.packet import Packet

from whad.exceptions import UnsupportedCapability
from whad.device import Device

from ..connector.base import ESB
from ..stack import ESBStack

class PTX(ESB):
    """
    Enhanced ShockBurst Primary Transmitter Role (PTX) implementation for compatible WHAD device.
    """
    def __init__(self, device: Device):
        """
        """
        super().__init__(device)

        # Check if device can modify its address and enter the PRX role
        self.__stack = ESBStack(self)
        self.__channel = 8
        self.__address = "11:22:33:44:55"

        self.__started = False
        if not self.can_set_node_address():
            raise UnsupportedCapability("SetNodeAddress")

        if not self.can_be_ptx():
            raise UnsupportedCapability("PrimaryTransmitterMode")
        self._enable_role()

    def _enable_role(self):
        self.set_node_address(self.__address)
        self.enable_ptx_mode(self.__channel)
        if self.__started:
            self.start()

    @property
    def stack(self) -> ESBStack:
        """Return the current ESB stack instance.
        """
        return self.__stack

    @property
    def channel(self) -> int:
        """Channel getter.

        :return: current channel
        :rtype: int
        """
        return self.__channel

    @channel.setter
    def channel(self, channel: int):
        """Change channel to `channel`.

        :param channel: New channel to use.
        :type channel: int
        """
        self.stop()
        self.__channel = channel
        self._enable_role()

    @property
    def address(self):
        """Return current address.

        :return: Current target address
        :rtype: str
        """
        return self.__address

    @address.setter
    def address(self, address: str):
        """Set target ESB address.

        :param address: Target ESB address
        :type address: str
        """
        self.stop()
        self.__address = address
        self._enable_role()

    def start(self):
        """Start PTX mode. A call to this method is required before calling
        :func:`send_data`.
        """
        super().start()
        self.__started = True

    def stop(self):
        """Stop PTX mode.
        """
        super().stop()
        self.__started = False

    def on_pdu(self, packet: Packet):
        """ESB packet callback.

        This function feeds the stack with the received packet.

        :param packet: Received ESB packet
        :type packet: :class:`scapy.packet.Packet`
        """
        self.__stack.on_pdu(packet)

    def send_data(self, data: bytes, waiting_ack: bool = False) -> bool:
        """Send an ESB payload.

        :param data: Data/payload to send
        :type data: bytes
        :param waiting_ack: If set to True, will wait for the acknowledgement
                            packet to be received
        :type waiting_ack: bool, optional
        :return: True if the packet has been correctly sent, False otherwise
        :rtype: bool
        """
        return self.__stack.ll.send_data(data, waiting_ack=waiting_ack)

    def synchronize(self, timeout: float = 10.0) -> bool:
        """Synchronize with the target ESB device.

        :param timeout: Timeout in seconds
        :type timeout: float
        :return: `True` if synchronization succeeded, `False` otherwise
        :rtype: bool
        """
        return self.__stack.ll.synchronize(timeout=timeout)



from whad.common.stack import LayerState

class Dot15d4Database(LayerState):
    """
    802.15.4 Generic Database of attributes.
    """
    def __init__(self):
        super().__init__()
        self.reset()

    def reset(self):
        """
        Reset the PIB database to its default value.
        """
        pass

    def get(self, attribute):
        """
        Read a given database attribute.
        """
        if hasattr(self, attribute):
            return getattr(self, attribute)
        return None

    def set(self, attribute, value):
        """
        Write a value to a given database attribute.
        """
        if hasattr(self, attribute):
            setattr(self, attribute, value)
            return True
        return False



"""Whad general replay tool

This tool implements a generic replay feature that can be used with compatible
domains. It takes a PCAP file in input, read packets and replay them through
a compatible WHAD device.
"""
import logging
from argparse import ArgumentParser
from dataclasses import fields, is_dataclass
from pkgutil import iter_modules
from inspect import getdoc
from html import escape
from importlib import import_module

# Scapy
from scapy.packet import Packet
from scapy.config import conf
from scapy.themes import BrightTheme

# Prompt
from prompt_toolkit import print_formatted_text, HTML

# Hexdump
from hexdump import hexdump

# WHAD
import whad
from whad.cli import CommandLineApp
from whad.common.replay import ReplayRole
from whad.exceptions import UnsupportedDomain, UnsupportedCapability
from whad.common.monitors import WiresharkMonitor, PcapWriterMonitor, WhadMonitor
from whad.common.pcap import PCAPReader

logger = logging.getLogger(__name__)

#logging.basicConfig(level=logging.DEBUG)

def list_implemented_replays():
    """Build a dictionnary of replays connector and configuration, by domain.
    """
    environment = {}

    # Iterate over modules
    for _, candidate_protocol,_ in iter_modules(whad.__path__):
        # If the module contains a replay connector and configuration,
        # store the associated classes in the environment dictionary
        try:
            module = import_module(f"whad.{candidate_protocol}.connector.replay")
            environment[candidate_protocol] = {
                "replay_class":module.Replay,
                "configuration_class":module.ReplayConfiguration
            }
        except ModuleNotFoundError:
            pass
    # return the environment dictionary
    return environment

def get_replay_parameters(configuration_class):
    """
    Extract all parameters from a replay configuration class, with their name
    and associated documentation.

    :param configuration_class: replay configuration class
    :return: dict containing parameters for a given configuration class
    """
    parameters = {}
    # Extract documentation of every field in the configuration class
    fields_configuration_documentation = {
        i.replace(":param ","").split(":")[0] : i.replace(":param ","").split(":")[1]
        for i in getdoc(configuration_class).split("\n")
        if i.startswith(":param ")
    }

    # Iterate over the fields of the configuration class
    for field in fields(configuration_class):

        # If the field is a dataclass, process subfields
        if is_dataclass(field.type):
            # Extract documentation of every subfields
            subfields_configuration_documentation = {
                i.replace(":param ","").split(":")[0] : i.replace(":param ","").split(":")[1]
                for i in getdoc(field.type).split("\n")
                if i.startswith(":param ")
            }

            # Populate parameters dict with subfields configuration
            for subfield in fields(field.type):
                parameters[f"{field.name}.{subfield.name}"] = (
                    subfield.type,
                    subfield.default,
                    field.type,
                    (
                        subfields_configuration_documentation[subfield.name]
                        if subfield.name in subfields_configuration_documentation
                        else None
                    )
                )
        # if the field is not a dataclass, process it
        else:
            # Populate parameters dict with field configuration
            parameters[field.name] = (
                field.type,
                field.default,
                None,
                (
                    fields_configuration_documentation[field.name]
                    if field.name in fields_configuration_documentation
                    else None
                )
            )
    return parameters


def build_configuration_from_args(environment, args):
    """
    Build replay configuration from arguments provided via argparse.

    :param environment: environment
    :type environment: dict
    :param args: arguments provided by user
    :type args: :class:`argparse.ArgumentParser`
    """
    configuration = environment[args.domain]["configuration_class"]()
    subfields = {}
    for parameter in environment[args.domain]["parameters"]:
        base_class = None

        base_class = environment[args.domain]["parameters"][parameter][2]

        if base_class is None:
            setattr(configuration,parameter,getattr(args,parameter))
        else:
            main, sub = parameter.split(".")
            if main not in subfields:
                subfields[main] = base_class()
            setattr(subfields[main], sub, getattr(args, parameter))

    for name, subfield in subfields.items():
        create = False
        for item in fields(subfield):
            if getattr(subfield, item.name) is not None:
                create = True
                break
        if create:
            setattr(configuration, name, subfield)
        else:
            setattr(configuration, name, None)
    return configuration

class WhadDomainSubParser(ArgumentParser):
    """
    Implements a Whad Domain subparser.
    """
    def warning(self, message):
        """Display a warning message in orange (if color is enabled)
        """
        print_formatted_text(
            HTML(f"<aaa fg=\"#e97f11\">/!\\ <b>{message}</b></aaa>")
        )

    def error(self, message):
        """Display an error message in red (if color is enabled)
        """
        print_formatted_text(
            HTML(f"<ansired>[!] <b>{message}</b></ansired>")
        )

class WhadReplayApp(CommandLineApp):
    """WHAD replay CLI application class.
    """

    class WhadReplayMonitor(WhadMonitor):
        """Custom packet monitor for wreplay.
        """
        def __init__(self, replay_app):
            super().__init__()
            self.__replay_app = replay_app

        def process_packet(self, packet):
            self.__replay_app.display(packet)

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD generic replay tool',
            interface=True,
            commands=False
        )

        self.add_argument(
            '--hide_metadata',
            dest='metadata',
            action="store_false",
            help='Hide packets metadata'
        )

        self.add_argument(
            '--format',
            dest='format',
            action="store",
            default='repr',
            choices=['repr', 'show', 'raw', 'hexdump'],
            help='Indicate format to display packet'
        )

        self.add_argument(
            '-p',
            '--pcapfile',
            dest='pcapfile',
            default=None,
            type=str,
            help='Input PCAP file'
        )

        self.add_argument(
            '-e',
            '--emitter',
            action='store_true',
            dest='emitter',
            default=False,
            help='Act as an emitter'
        )

        self.add_argument(
            '-r',
            '--receiver',
            action='store_true',
            dest='receiver',
            default=False,
            help='Act as a receiver'
        )

        self.add_argument(
            '-o',
            '--output',
            dest='output',
            default=None,
            type=str,
            help='Output PCAP file'
        )

        self.add_argument(
            '-w',
            '--wireshark',
            dest='wireshark',
            action='store_true',
            help='Enable wireshark monitoring'
        )

        self.add_argument(
            '-s',
            '--start',
            dest='start_pos',
            default=0,
            type=int,
            help='Specify the start position in the PCAP file'
        )

        self.add_argument(
            '-z',
            '--end_pos',
            dest='stop_pos',
            default=-1,
            type=int,
            help='Specify the stop position in the PCAP file'
        )

        self.add_argument(
            '-x',
            '--exclude',
            dest='exclude',
            action='append',
            help='Ignore a specific packet',
            type=int,
            default=[]
        )

        self.add_argument(
            '-d',
            '--delay',
            dest='offset',
            default=100,
            type=int,
            help='Additional delay between packets, in milliseconds'
        )

        subparsers = self.add_subparsers(
            required=True,
            dest="domain",
            parser_class=WhadDomainSubParser,
            help='Domain in use'
        )

        self.build_subparsers(subparsers)


    def build_monitor(self) -> WhadReplayMonitor:
        """Create our own wrapper to monitor packets in the terminal.
        """
        return WhadReplayApp.WhadReplayMonitor(self)


    def display(self, pkt):
        """
        Display a packet according to the selected format.

        Four main types of formats can be used:
            * repr: scapy packet repr method (default)
            * show: scapy show method, "field" representation
            * hexdump: hexdump representation of the packet content
            * raw: raw received bytes

        :param  pkt:        Received Signal Strength Indicator
        :type   pkt:        :class:`scapy.packet.packet`
        """
        if isinstance(pkt, Packet):

            metadata = ""
            if hasattr(pkt, "metadata") and self.args.metadata:
                metadata = repr(pkt.metadata)

            # Process scapy show method format
            if self.args.format == "show":
                print_formatted_text(HTML(
                    f"<b><ansipurple>{metadata}</ansipurple></b>"
                ))
                pkt.show()

                if hasattr(pkt, "decrypted"):
                    print_formatted_text(HTML(
                        "<ansicyan>[i] Decrypted payload:</ansicyan>"
                    ))
                    pkt.decrypted.show()

            # Process raw bytes format
            elif self.args.format == "raw":
                print_formatted_text(HTML(
                    f"<b><ansipurple>{metadata}</ansipurple></b> {bytes(pkt).hex()}"
                ))

                if hasattr(pkt, "decrypted"):
                    print_formatted_text(HTML(
                        f"<ansicyan>[i] Decrypted payload:</ansicyan> {bytes(pkt.decrypted).hex()}"
                    ))

            # Process hexdump format
            elif self.args.format == "hexdump":
                print_formatted_text(HTML(
                    f"<b><ansipurple>{metadata}</ansipurple></b>"
                ))

                pkt_bytes = escape(hexdump(bytes(pkt), result="return"))
                print_formatted_text(HTML(f"<i>{pkt_bytes}</i>"))

                if hasattr(pkt, "decrypted"):
                    print_formatted_text(
                        HTML(
                            "<ansicyan>[i] Decrypted payload:</ansicyan>"
                        )
                    )

                    pkt_decrypted = escape(hexdump(bytes(pkt.decrypted), result="return"))
                    print_formatted_text(HTML(f"<i>{pkt_decrypted}</i>"))

            # Process scapy repr format
            else:
                print_formatted_text(HTML(
                    f"<b><ansipurple>{metadata}</ansipurple></b>"
                ))
                print(repr(pkt))
                if hasattr(pkt, "decrypted"):
                    print_formatted_text(
                        HTML("<ansicyan>[i] Decrypted payload:</ansicyan>")
                    )
                    print(repr(pkt.decrypted))
            print()
        # If it is not a packet, use repr method
        else:
            print(repr(pkt))

    def display_event(self, event):
        """Display an event generated from a replay.
        """
        event_details = "("+event.message +")" if event.message is not None else ""
        print_formatted_text(HTML(
            f"<ansicyan>[i] event: <b>{event.name}</b></ansicyan> {event_details}"
        ))

    def pre_run(self):
        """Pre-run operations: configure scapy theme.
        """
        super().pre_run()
        # If no color is not selected, configure scapy color theme
        if not self.args.nocolor:
            conf.color_theme = BrightTheme()

    def run(self):
        monitors = []
        replay = None
        # Launch pre-run tasks
        self.pre_run()
        try:
            # We need to have at least one role selected
            role = 0
            if self.args.emitter:
                role |= ReplayRole.EMITTER
            if self.args.receiver:
                role |= ReplayRole.RECEIVER

            # If role==0, no role selected
            if role == 0:
                self.error("You need to specify a role with --emitter and/or --receiver.")
                return

            # We also must have an input pcap file selected
            if self.args.pcapfile is None:
                self.error("No input PCAP file specified !")
                return

            # We need to have an interface specified
            if self.interface is not None:
                # We need to have a domain specified
                if self.args.domain is not None:
                    # Parse the arguments to populate a replay configuration
                    configuration = build_configuration_from_args(self.environment, self.args)

                    # Make sure we have a target set
                    if configuration.target is not None:
                        # Generate a replay based on the selected domain
                        replay = self.environment[self.args.domain]["replay_class"](
                            self.interface,
                            self.args.pcapfile,
                            role=role
                        )

                        # Add our own monitor for CLI reporting
                        cli_monitor = self.build_monitor()
                        cli_monitor.attach(replay)
                        cli_monitor.start()
                        monitors.append(cli_monitor)

                        # If output parameter is selected, add a PCAP Writer monitor
                        if self.args.output is not None:
                            monitor_pcap = PcapWriterMonitor(self.args.output)
                            monitor_pcap.attach(replay)
                            monitor_pcap.start()
                            monitors.append(monitor_pcap)

                        # If wireshark parameter is selected, add a WiresharkMonitor
                        if self.args.wireshark:
                            monitor_wireshark = WiresharkMonitor()
                            monitor_wireshark.attach(replay)
                            monitor_wireshark.start()
                            monitors.append(monitor_wireshark)

                        # Prepare the replay instance
                        if replay.prepare(configuration):
                            # Now we can feed our replay instance with packets
                            if self.args.stop_pos >= 0:
                                count = self.args.stop_pos - self.args.start_pos + 1
                            else:
                                count = None

                            # PCAPReader will send back packets in a timely manner,
                            # according to PCAP timestamps.
                            reader = PCAPReader(self.args.pcapfile)
                            for packet in reader.packets(start=self.args.start_pos, count=count,
                                                        offset=self.args.offset/1000.,
                                                        exclude=self.args.exclude):
                                replay.send_packet(packet)

                        # Stop our replay instance
                        replay.stop()
                        replay.close()

                        # Close all monitors
                        for monitor in monitors:
                            monitor.close()
                    else:
                        self.error("You must specify a target BD address with option --target.")
                else:
                    self.error("You need to specify a domain.")
            else:
                self.error("You need to specify an interface with option --interface.")


        except UnsupportedDomain:
            self.error(f"WHAD device doesn\'t support selected domain ({self.args.domain})")

        except UnsupportedCapability as unsupported_capability:
            self.error((f"WHAD device doesn\'t support selected capability "
                       f"({unsupported_capability.capability})"))

        except KeyboardInterrupt:
            self.warning("replay stopped (CTRL-C)")
            replay.stop()
            replay.close()
            for monitor in monitors:
                monitor.close()

    def build_subparsers(self, subparsers):
        """
        Generate the subparsers argument according to the environment.
        """

        # List every domain implementing a replay
        self.environment = list_implemented_replays()

        # Iterate over domain, and get the associated replay parameters
        for name, domain in self.environment.items():
            domain["parameters"] = get_replay_parameters(
                domain["configuration_class"]
            )

            domain["subparser"] = subparsers.add_parser(
                name,
                description=f"WHAD {name.capitalize()} Replay tool"
            )
            # Iterate over every parameters, and add arguments to subparsers
            for (
                    parameter_name,
                    (parameter_type, parameter_default, parameter_base_class, parameter_help)
                ) in domain["parameters"].items():

                dest = parameter_name
                # If parameter is based on a dataclass, process a subparameter
                if parameter_base_class is not None:
                    _, parameter_name = parameter_name.split(".")

                # Process parameter help and shortname
                if parameter_help is not None and "(" in parameter_help:
                    parameter_shortnames = [
                        f"-{i}" for i in
                        parameter_help.split("(")[1].replace(")","").split(",")
                    ]
                    parameter_help = parameter_help.split("(")[0]
                else:
                    parameter_shortnames = []


                # Process parameter type
                if parameter_type is not bool:
                    # If we got an int
                    if parameter_type is int:
                        # allow to provide hex arguments
                        parameter_type = lambda x: int(x,0)
                    # If we got a list, it is a list of string
                    if parameter_type is list:
                        parameter_default=[]
                        parameter_type=str
                        action = "append"
                    elif parameter_type is bytes:
                        parameter_type = bytes.fromhex
                        action = "store"
                    else:
                        action = "store"
                    # Add non-boolean argument to corresponding subparser
                    domain["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        default=parameter_default,
                        action=action,
                        type=parameter_type,
                        dest=dest,
                        help=parameter_help
                    )
                else:
                    # Add boolean argument to corresponding subparser
                    domain["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        action="store_true",
                        dest=dest,
                        help=parameter_help
                    )


def whadreplay_main():
    """Launcher for wreplay.
    """
    app = WhadReplayApp()
    app.run()



"""WHAD Protocol Generic CommandResult messages abstraction layer.
"""
from whad.protocol.whad_pb2 import Message
from whad.protocol.generic_pb2 import ResultCode
from whad.hub.message import HubMessage,pb_bind
from whad.hub.generic import Generic

@pb_bind(Generic, 'cmd_result', 1)
class CommandResult(HubMessage):
    """CommandResult message class

    Provides a parsing method (`parse()`) to parse a CommandResult protocol
    buffers message into an object, based on its result code.
    """

    ERROR = ResultCode.ERROR
    SUCCESS = ResultCode.SUCCESS
    PARAMETER_ERROR = ResultCode.PARAMETER_ERROR
    DISCONNECTED = ResultCode.DISCONNECTED
    WRONG_MODE = ResultCode.WRONG_MODE
    UNSUPPORTED_DOMAIN = ResultCode.UNSUPPORTED_DOMAIN
    BUSY = ResultCode.BUSY


    def __init__(self, result_code: int = ResultCode.SUCCESS, message: Message = None):
        """Create a new CommandResult message with the associated protocol
        version number.
        """
        super().__init__(message=message)

        # If ProtoBuf message is provided, extract result code from message
        if message is not None:
            self.__result_code = message.generic.cmd_result.result

        # If result_code is provided, override result code value
        if result_code is not None:
            self.result_code = result_code

    @property
    def result_code(self) -> int:
        """Return the CommandResult result code.
        """
        return self.__result_code

    @result_code.setter
    def result_code(self, result_code: int):
        self.__result_code = result_code

    def serialize(self) -> bytes:
        """Serialize this message using Protobuf serialization routines.
        """
        self.message.generic.cmd_result.result = self.__result_code
        return super().serialize()

    @staticmethod
    def parse(version: int, msg):
        """Parse a protocol buffer message containing a CommandResult object
        into the corresponding generic message class.
        """
        if msg.generic.cmd_result.result == CommandResult.ERROR:
            return Generic.bound('cmd_result_error', version)(message=msg)
        elif msg.generic.cmd_result.result == CommandResult.SUCCESS:
            return Generic.bound('cmd_result_success', version)(message=msg)
        elif msg.generic.cmd_result.result == CommandResult.PARAMETER_ERROR:
            return Generic.bound('cmd_result_param_error', version)(message=msg)
        elif msg.generic.cmd_result.result == CommandResult.DISCONNECTED:
            return Generic.bound('cmd_result_disconnected', version)(message=msg)
        elif msg.generic.cmd_result.result == CommandResult.WRONG_MODE:
            return Generic.bound('cmd_result_wrong_mode', version)(message=msg)
        elif msg.generic.cmd_result.result == CommandResult.UNSUPPORTED_DOMAIN:
            return Generic.bound('cmd_result_unsupported_domain', version)(message=msg)
        elif msg.generic.cmd_result.result == CommandResult.BUSY:
            return Generic.bound('cmd_result_busy', version)(message=msg)


@pb_bind(Generic, 'cmd_result_error', 1)
class Error(CommandResult):
    """Generic error message.
    """

    def __init__(self, message: Message = None):
        """Create a CommandResult message with result code ERROR.
        """
        super().__init__(result_code=CommandResult.ERROR, message=message)

@pb_bind(Generic, 'cmd_result_success', 1)
class Success(CommandResult):
    """Generic success message.
    """

    def __init__(self, message: Message = None):
        """Create a CommandResult message with result code SUCCESS.
        """
        super().__init__(result_code=CommandResult.SUCCESS, message=message)

@pb_bind(Generic, 'cmd_result_param_error', 1)
class ParameterError(CommandResult):
    """Generic parameter error message.
    """

    def __init__(self, message: Message = None):
        """Create a CommandResult message with result code PARAMETER_ERROR.
        """
        super().__init__(result_code=CommandResult.PARAMETER_ERROR, message=message)

@pb_bind(Generic, 'cmd_result_disconnected', 1)
class Disconnected(CommandResult):
    """Generic disconnected message.
    """

    def __init__(self, message: Message = None):
        """Create a CommandResult message with result code DISCONNECTED.
        """
        super().__init__(result_code=CommandResult.DISCONNECTED, message=message)

@pb_bind(Generic, 'cmd_result_wrong_mode', 1)
class WrongMode(CommandResult):
    """Generic wrong mode message.
    """

    def __init__(self, message: Message = None):
        """Create a CommandResult message with result code WRONG_MODE.
        """
        super().__init__(result_code=CommandResult.WRONG_MODE, message=message)

@pb_bind(Generic, 'cmd_result_unsupported_domain', 1)
class UnsupportedDomain(CommandResult):
    """Generic unsupported domain error message.
    """

    def __init__(self, message: Message = None):
        """Create a CommandResult message with result code UNSUPPORTED_DOMAIN.
        """
        super().__init__(result_code=CommandResult.UNSUPPORTED_DOMAIN, message=message)

@pb_bind(Generic, 'cmd_result_busy', 1)
class Busy(CommandResult):
    """Generic busy message.
    """

    def __init__(self, message: Message = None):
        """Create a CommandResult message with result code BUSY.
        """
        super().__init__(result_code=CommandResult.BUSY, message=message)
